import logging
import re
from collections import defaultdict
from typing import Dict, List, cast

import torch
from torch import Tensor, nn
from tqdm.autonotebook import tqdm

import fusion_bench.method.regmean.utils as regmean_utils
from fusion_bench import BaseAlgorithm, auto_register_config
from fusion_bench.mixins import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool

log = logging.getLogger(__name__)


def reduce_non_diagonal_elements(
    regmean_weights: torch.Tensor, reduce_non_diagonal_ratio: float
):
    """
    reduce the non-diagonal elements in regmean_weights
    :param regmean_weights: Tensor, shape (hidden_dim, hidden_dim), input regmean weights
    :param reduce_non_diagonal_ratio: float, reduce non-diagonal elements in regmean weights by multiplying this scalar
    :return:
    """
    # diagonal matrix with (1 - reduce_non_diagonal_ratio) as elements
    diag_weights = torch.diag(
        torch.ones(regmean_weights.shape[0]) - reduce_non_diagonal_ratio
    ).to(regmean_weights.device)
    # matrix with reduce_non_diagonal_ratio as elements
    non_diag_weights = torch.zeros_like(diag_weights).fill_(reduce_non_diagonal_ratio)
    # diagonal elements are unchanged, while non-diagonal elements are multiplied by reduce_non_diagonal_ratio
    return regmean_weights * (diag_weights + non_diag_weights)


def regmean_params_merge(
    param_weight_list: List[Tensor],
    param_regmean_list: List[Tensor],
    reduce_non_diagonal_ratio: float = 1.0,
    weight_transpose: bool = True,
    module_name: str = "",
    device="cpu",
):
    # two lists with length num_models_to_merge
    param_multiplied_results, module_regmean_weights_list = [], []
    for model_idx, module_regmean_weights in enumerate(param_regmean_list):
        # reduce non-diagonal elements
        module_regmean_weights = reduce_non_diagonal_elements(
            regmean_weights=module_regmean_weights,
            reduce_non_diagonal_ratio=reduce_non_diagonal_ratio,
        )
        module_regmean_weights_list.append(module_regmean_weights)

        model_to_merge_param = param_weight_list[model_idx]
        # since the weight shape of Linear module is (output_size, input_size), we need to transpose it
        param_multiplied_results.append(
            torch.matmul(
                module_regmean_weights,
                (
                    model_to_merge_param.transpose(0, 1)
                    if weight_transpose
                    else model_to_merge_param
                ),
            )
        )

    # sum up module_regmean_weights and param_multiplied_results over all individual models
    sum_module_regmean_weights = sum(module_regmean_weights_list)
    sum_param_multiplied_results = sum(param_multiplied_results)

    # get the inverse matrix
    inv_sum_module_regmean_weights = torch.inverse(sum_module_regmean_weights)
    # merge parameters with regmean
    merged_param = torch.matmul(
        inv_sum_module_regmean_weights, sum_param_multiplied_results
    )
    # transpose to the original shape of "weight" in Linear module
    merged_param = merged_param.transpose(0, 1) if weight_transpose else merged_param

    return merged_param


def merging_with_regmean_weights(
    models_to_merge_param_dict: dict,
    models_to_merge_regmean_weights_list: list,
    reduce_non_diagonal_ratio: float = 1.0,
    weight_transpose: bool = True,
):
    """
    merge parameters of different models with computed regmean weights

    Asrgs:
        models_to_merge_param_dict: dict, dictionary of list, where key is the parameter name,
            value is a list of the corresponding parameters of all the models that need to be merged
        models_to_merge_regmean_weights_list: list, list of dictionaries with length len(models_to_merge),
            each dictionary records the regmean weights (matrix) of parameters for each model that needs to be merged, key is module name
        reduce_non_diagonal_ratio: float, reduce non-diagonal elements in regmean weights by multiplying this scalar

    Returns:
        dict: merged model parameters
    """
    # dict, dictionary of model parameters
    merged_params = {}

    for param_name, param_value_list in models_to_merge_param_dict.items():
        merged_by_regmean = False
        # only perform regmean merging on the "weight" parameter of Linear module
        if param_name.endswith(".weight"):
            module_name = param_name[: -len(".weight")]
            if module_name in models_to_merge_regmean_weights_list[0].keys():
                # two lists with length num_models_to_merge
                module_regmean_weights_list = []
                for model_idx, model_to_merge_regmean_weights in enumerate(
                    models_to_merge_regmean_weights_list
                ):
                    device = param_value_list[model_idx].device

                    # Tensor, shape (hidden_dim, hidden_dim)
                    module_regmean_weights = model_to_merge_regmean_weights[
                        module_name
                    ].to(device)
                    module_regmean_weights_list.append(module_regmean_weights)

                merged_params[param_name] = regmean_params_merge(
                    param_weight_list=param_value_list,
                    param_regmean_list=module_regmean_weights_list,
                    reduce_non_diagonal_ratio=reduce_non_diagonal_ratio,
                    weight_transpose=weight_transpose,
                    module_name=module_name,
                    device=device,
                )

                merged_by_regmean = True
        # use average merging for parameters whose names are not end with ".weight" or not in Linear module
        if not merged_by_regmean:
            merged_params[param_name] = torch.stack(param_value_list, dim=0).mean(dim=0)

    return merged_params


@auto_register_config
class RegMeanAlgorithmPlusPlus(
    SimpleProfilerMixin,
    BaseAlgorithm,
):
    _include_module_type = [nn.Linear]

    def __init__(
        self,
        *,
        num_regmean_examples: int,
        exclude_param_names_regex: list,
        reduce_non_diagonal_ratio: float,
        weight_transpose: bool,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.num_regmean_examples = num_regmean_examples
        self.exclude_param_names_regex = exclude_param_names_regex
        self.reduce_non_diagonal_ratio = reduce_non_diagonal_ratio
        self.weight_transpose = weight_transpose

    def run(self, modelpool: BaseModelPool, **kwargs):
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(modelpool)
        self.modelpool = modelpool
        device = "cuda:0" if torch.cuda.is_available() else "cpu"
        models_to_merge_dict = {
            name: model.to(device) for name, model in modelpool.named_models()
        }
        self.on_regmean_start()

        # initialize the merged models as the pretrained model
        merged_model = modelpool.load_pretrained_model().to(device)
        merged_params_dict = {}

        # 1. merge embedding layer
        merged_embedding_dict = self.merge_embedding_layer(
            models_to_merge_dict=models_to_merge_dict
        )
        merged_model.load_state_dict(merged_embedding_dict, strict=False)

        with torch.no_grad():
            # 1.1. compute input for the first layer
            with (
                self.profile("merging models"),
                self.profile("computing first layer input"),
            ):
                batches_input_dict = defaultdict(list)
                for name in tqdm(
                    models_to_merge_dict.keys(), desc="computing input for first layer"
                ):
                    dataset = modelpool.load_train_dataset(name)

                    batches_input_dict[name] = self.get_input_for_first_layer(
                        merged_model, dataset
                    )

            # 2. iteratively merge layer by layer with regmean algorithm
            backbone_layers = self.get_layers(merged_model)
            num_layers = len(backbone_layers)

            models_to_merge_layers_dict = defaultdict(list)
            for name, model in models_to_merge_dict.items():
                models_to_merge_layers_dict[name] = self.get_layers(model)

            param_names_to_merge = None
            for layer_idx, backbone_layer in tqdm(
                enumerate(backbone_layers), desc="merging layers", total=num_layers
            ):
                # dictionary of list, where key is the parameter name,
                # value is a list of the corresponding parameters of all the models that need to be merged
                models_to_merge_param_dict = defaultdict(list)

                # list of dictionaries with length len(models_to_merge),
                # each dictionary records the regmean weights (matrix) of parameters for each model that needs to be merged
                models_to_merge_regmean_weights_list = []

                for name, layers_to_merge in models_to_merge_layers_dict.items():
                    layer_to_merge = layers_to_merge[layer_idx]
                    param_dict = layer_to_merge.state_dict()

                    # exclude parameter whose name matches element in exclude_param_names_regex
                    if param_names_to_merge is None:
                        param_names_to_merge = regmean_utils.get_param_names_to_merge(
                            input_param_names=list(param_dict.keys()),
                            exclude_param_names_regex=self.config.get(
                                "exclude_param_names_regex", []
                            ),
                        )

                    for param_name in param_names_to_merge:
                        models_to_merge_param_dict[param_name].append(
                            param_dict[param_name]
                        )

                    linear_modules_to_merge = regmean_utils.get_modules_to_merge(
                        model=layer_to_merge,
                        include_module_types=self._include_module_type,
                    )
                    assert (
                        len(linear_modules_to_merge) > 0
                    ), "No linear modules to merge"

                    # 2.1. compute regmean weights for each model
                    with (
                        self.profile("merging models"),
                        self.profile("computing regmean weights"),
                    ):
                        regmean_weights = self.get_regmean_weights(
                            name,
                            layer_to_merge,
                            batches_input=batches_input_dict[name],
                            linear_modules_to_merge=linear_modules_to_merge,
                        )

                        module_subset = regmean_utils.get_param_names_to_merge(
                            input_param_names=list(param_dict.keys()),
                            exclude_param_names_regex=self.exclude_param_names_regex,
                        )
                        module_subset = [
                            name.replace(".weight", "").replace(".bias", "")
                            for name in module_subset
                        ]
                        module_subset = list(set(module_subset))
                        regmean_weights = {
                            module_name: regmean_weights[module_name]
                            for module_name in module_subset
                            if module_name in regmean_weights
                        }

                        models_to_merge_regmean_weights_list.append(regmean_weights)

                # 2.2. merge parameters with regmean weights
                with self.profile("merging models"):
                    # merging with regmean weights
                    merged_layer_params = merging_with_regmean_weights(
                        models_to_merge_param_dict=models_to_merge_param_dict,
                        models_to_merge_regmean_weights_list=models_to_merge_regmean_weights_list,
                        reduce_non_diagonal_ratio=self.reduce_non_diagonal_ratio,
                        weight_transpose=self.config.get("weight_transpose", True),
                    )

                    merged_params_dict = self.update_merged_params_dict(
                        merged_params_dict=merged_params_dict,
                        new_merged_params=merged_layer_params,
                        layer_idx=layer_idx,
                    )

                # 2.3. compute input for the next layer
                with (
                    self.profile("merging models"),
                    self.profile("forwarding next layer"),
                ):
                    if layer_idx < num_layers - 1:
                        backbone_layer.load_state_dict(
                            merged_layer_params, strict=False
                        )
                        batches_output_dict = defaultdict(list)
                        for name in models_to_merge_dict.keys():
                            batches_output_dict[name] = self.layer_batches_forward(
                                backbone_layer, batches_input_dict[name]
                            )
                        batches_input_dict = batches_output_dict

            # 3. load state dict to the merged model
            merged_model.load_state_dict(merged_params_dict, strict=False)

        self.print_profile_summary()
        return merged_model

    def merge_embedding_layer(self, models_to_merge_dict: Dict[str, nn.Module]):
        """
        Merge the embedding layer of the model with the merged model.
        This method should be implemented in subclasses if needed.
        """
        raise NotImplementedError()

    def get_input_for_first_layer(self, model: nn.Module, train_dataset):
        raise NotImplementedError

    def get_layers(self, model: nn.Module):
        raise NotImplementedError

    def update_merged_params_dict(
        self, merged_params_dict, new_merged_params, layer_idx
    ):
        raise NotImplementedError

    def layer_batches_forward(self, layer: nn.Module, batches_input: List[Tensor]):
        raise NotImplementedError

    def on_regmean_start(self):
        pass

    def get_regmean_weights(
        self,
        model_name: str,
        layer: nn.Module,
        batches_input: List[Tensor],
        linear_modules_to_merge: Dict[str, nn.Module],
    ):
        raise NotImplementedError



import logging
from copy import deepcopy
from typing import Literal, Optional, Union

from datasets import load_dataset
from lightning.fabric.utilities import rank_zero_only
from omegaconf import DictConfig, open_dict
from torch import nn
from torch.utils.data import Dataset
from transformers import CLIPModel, CLIPProcessor, CLIPVisionModel
from typing_extensions import override

from fusion_bench.utils import instantiate, timeit_context
from fusion_bench.utils.modelscope import resolve_repo_path

from ..base_pool import BaseModelPool

log = logging.getLogger(__name__)


class CLIPVisionModelPool(BaseModelPool):
    """
    A model pool for managing Hugging Face's CLIP Vision models.

    This class extends the base `ModelPool` class and overrides its methods to handle
    the specifics of the CLIP Vision models provided by the Hugging Face Transformers library.
    """

    _config_mapping = BaseModelPool._config_mapping | {
        "_processor": "processor",
        "_platform": "hf",
    }

    def __init__(
        self,
        models: DictConfig,
        *,
        processor: Optional[DictConfig] = None,
        platform: Literal["hf", "huggingface", "modelscope"] = "hf",
        **kwargs,
    ):
        super().__init__(models, **kwargs)
        self._processor = processor
        self._platform = platform

    def load_processor(self, *args, **kwargs) -> CLIPProcessor:
        assert self._processor is not None, "Processor is not defined in the config"
        if isinstance(self._processor, str):
            if rank_zero_only.rank == 0:
                log.info(f"Loading `transformers.CLIPProcessor`: {self._processor}")
            repo_path = resolve_repo_path(
                repo_id=self._processor, repo_type="model", platform=self._platform
            )
            processor = CLIPProcessor.from_pretrained(repo_path, *args, **kwargs)
        else:
            processor = instantiate(self._processor, *args, **kwargs)
        return processor

    def load_clip_model(self, model_name: str, *args, **kwargs) -> CLIPModel:
        model_config = self._models[model_name]

        if isinstance(model_config, str):
            if rank_zero_only.rank == 0:
                log.info(f"Loading `transformers.CLIPModel`: {model_config}")
            repo_path = resolve_repo_path(
                repo_id=model_config, repo_type="model", platform=self._platform
            )
            clip_model = CLIPModel.from_pretrained(repo_path, *args, **kwargs)
            return clip_model
        else:
            assert isinstance(
                model_config, DictConfig
            ), "Model config must be a DictConfig"
            model_config = deepcopy(model_config)
            with open_dict(model_config):
                model_config._target_ = "transformers.CLIPModel.from_pretrained"
            clip_model = instantiate(model_config, *args, **kwargs)
            return clip_model

    @override
    def save_model(self, model: CLIPVisionModel, path: str):
        """
        Save a CLIP Vision model to the given path.

        Args:
            model (CLIPVisionModel): The model to save.
            path (str): The path to save the model to.
        """
        with timeit_context(f'Saving clip vision model to "{path}"'):
            model.save_pretrained(path)

    def load_model(
        self, model_name_or_config: Union[str, DictConfig], *args, **kwargs
    ) -> CLIPVisionModel:
        """
        This method is used to load a CLIPVisionModel from the model pool.

        Example configuration could be:

        ```yaml
        models:
            cifar10: tanganke/clip-vit-base-patch32_cifar10
            sun397: tanganke/clip-vit-base-patch32_sun397
            stanford-cars: tanganke/clip-vit-base-patch32_stanford-cars
        ```

        Args:
            model_name_or_config (Union[str, DictConfig]): The name of the model or the model configuration.

        Returns:
            CLIPVisionModel: The loaded CLIPVisionModel.
        """
        if (
            isinstance(model_name_or_config, str)
            and model_name_or_config in self._models
        ):
            model = self._models[model_name_or_config]
            if isinstance(model, str):
                if rank_zero_only.rank == 0:
                    log.info(f"Loading `transformers.CLIPVisionModel`: {model}")
                repo_path = resolve_repo_path(
                    model, repo_type="model", platform=self._platform
                )
                return CLIPVisionModel.from_pretrained(repo_path, *args, **kwargs)
            if isinstance(model, nn.Module):
                if rank_zero_only.rank == 0:
                    log.info(f"Returning existing model: {model}")
                return model
        else:
            # If the model is not a string, we use the default load_model method
            return super().load_model(model_name_or_config, *args, **kwargs)

    def load_train_dataset(self, dataset_name: str, *args, **kwargs):
        dataset_config = self._train_datasets[dataset_name]
        if isinstance(dataset_config, str):
            if rank_zero_only.rank == 0:
                log.info(
                    f"Loading train dataset using `datasets.load_dataset`: {dataset_config}"
                )
            dataset = self._load_dataset(dataset_config, split="train")
        else:
            dataset = super().load_train_dataset(dataset_name, *args, **kwargs)
        return dataset

    def load_val_dataset(self, dataset_name: str, *args, **kwargs):
        dataset_config = self._val_datasets[dataset_name]
        if isinstance(dataset_config, str):
            if rank_zero_only.rank == 0:
                log.info(
                    f"Loading validation dataset using `datasets.load_dataset`: {dataset_config}"
                )
            dataset = self._load_dataset(dataset_config, split="validation")
        else:
            dataset = super().load_val_dataset(dataset_name, *args, **kwargs)
        return dataset

    def load_test_dataset(self, dataset_name: str, *args, **kwargs):
        dataset_config = self._test_datasets[dataset_name]
        if isinstance(dataset_config, str):
            if rank_zero_only.rank == 0:
                log.info(
                    f"Loading test dataset using `datasets.load_dataset`: {dataset_config}"
                )
            dataset = self._load_dataset(dataset_config, split="test")
        else:
            dataset = super().load_test_dataset(dataset_name, *args, **kwargs)
        return dataset

    def _load_dataset(self, name: str, split: str):
        """
        Load a dataset by its name and split.

        Args:
            dataset_name (str): The name of the dataset.
            split (str): The split of the dataset to load (e.g., "train", "validation", "test").

        Returns:
            Dataset: The loaded dataset.
        """
        datset_dir = resolve_repo_path(
            name, repo_type="dataset", platform=self._platform
        )
        dataset = load_dataset(datset_dir, split=split)
        return dataset



import logging
import os
from typing import List

log = logging.getLogger(__name__)


def path_is_dir_and_not_empty(path: str):
    if path is None:
        return False
    return os.path.isdir(path) and len(os.listdir(path)) > 0


def listdir_fullpath(dir: str) -> List[str]:
    """list directory `dir`, return fullpaths

    Args:
        dir (str): directory name

    Returns:
        List[str]: a list of fullpaths
    """
    assert os.path.isdir(dir), "Argument 'dir' must be a Directory"
    names = os.listdir(dir)
    return [os.path.join(dir, name) for name in names]


def create_symlink(src_dir: str, dst_dir: str, link_name: str = None):
    """
    Creates a symbolic link from src_dir to dst_dir.

    Args:
        src_dir (str): The source directory to link to.
        dst_dir (str): The destination directory where the symlink will be created.
        link_name (str, optional): The name of the symlink. If None, uses the basename of src_dir.

    Raises:
        OSError: If the symbolic link creation fails.
        ValueError: If src_dir does not exist or is not a directory.
    """
    if not os.path.exists(src_dir):
        raise ValueError(f"Source directory does not exist: {src_dir}")

    if not os.path.isdir(src_dir):
        raise ValueError(f"Source path is not a directory: {src_dir}")

    # Avoid creating symlink if source and destination are the same
    if os.path.abspath(src_dir) == os.path.abspath(dst_dir):
        log.warning(
            "Source and destination directories are the same, skipping symlink creation"
        )
        return

    # Create destination directory if it doesn't exist
    os.makedirs(dst_dir, exist_ok=True)

    # Determine link name
    if link_name is None:
        link_name = os.path.basename(src_dir)

    link_path = os.path.join(dst_dir, link_name)

    try:
        # if the system is windows, use the `mklink` command in "CMD" to create the symlink
        if os.name == "nt":
            os.system(
                f"mklink /J {os.path.abspath(link_path)} {os.path.abspath(src_dir)}"
            )
        else:
            os.symlink(
                src_dir,
                link_path,
                target_is_directory=True,
            )
        log.info(f"Created symbolic link: {link_path} -> {src_dir}")
    except OSError as e:
        log.warning(f"Failed to create symbolic link: {e}")
        raise



import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


def visualize_model_inputs(input_ids, attention_mask, labels, tokenizer=None):
    """
    Visualize model inputs: attention mask, labels and input_ids

    Parameters:
    -----------
    attention_mask: numpy array or tensor
        The attention mask array
    labels: numpy array or tensor
        The labels array
    input_ids: numpy array or tensor
        The input ids array
    tokenizer: optional
        The tokenizer object to decode input_ids
    """

    # Convert inputs to numpy if they're tensors
    attention_mask = np.array(attention_mask)
    labels = np.array(labels)
    input_ids = np.array(input_ids)

    # Create figure with 3 subplots
    fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(15, 10))

    # Plot attention mask
    sns.heatmap(attention_mask.reshape(1, -1), ax=ax1, cmap="Blues", cbar=True)
    ax1.set_title("**Attention Mask**")
    ax1.set_ylabel("Sequence")

    # Plot labels
    sns.heatmap(labels.reshape(1, -1), ax=ax2, cmap="Reds", cbar=True)
    ax2.set_title("**Labels**")
    ax2.set_ylabel("Sequence")

    # Plot input_ids
    sns.heatmap(input_ids.reshape(1, -1), ax=ax3, cmap="Greens", cbar=True)
    ax3.set_title("**Input IDs**")
    ax3.set_ylabel("Sequence")

    # If tokenizer is provided, add decoded tokens as x-axis labels
    if tokenizer:
        decoded_tokens = [tokenizer.decode(token_id) for token_id in input_ids]
        ax3.set_xticks(np.arange(len(decoded_tokens)) + 0.5)
        ax3.set_xticklabels(decoded_tokens, rotation=45, ha="right")

    plt.tight_layout()
    return fig



"""
For example:

Fine-tune CLIP-ViT-B/32:

```bash
fusion_bench \
    method=clip_finetune \
    modelpool=clip-vit-base-patch32_mtl \
    taskpool=dummy
```

Fine-tune CLIP-ViT-L/14 on eight GPUs with a per-device per-task batch size of 2.

```bash
fusion_bench \
    fabric.devices=8 \
    method=clip_finetune \
        method.batch_size=2 \
    modelpool=clip-vit-base-patch32_mtl \
        modelpool.models.0.path=openai/clip-vit-large-patch14 \
    taskpool=dummy
```
"""

import os
from typing import Optional, Tuple

import lightning as L
import torch
from omegaconf import OmegaConf
from peft import LoraConfig, PeftModel, get_peft_model
from peft.tuners.lora import LoraLayer
from safetensors.torch import save_file
from torch import nn
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from transformers import CLIPModel, CLIPProcessor, CLIPVisionModel
from transformers.models.clip.modeling_clip import CLIPVisionTransformer

from fusion_bench import print_parameters
from fusion_bench.compat.method import ModelFusionAlgorithm
from fusion_bench.compat.modelpool import to_modelpool
from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.models.hf_clip import HFCLIPClassifier
from fusion_bench.models.linearized.linearized_model_utils import LinearizedModelWraper
from fusion_bench.utils.data import InfiniteDataLoader


def _get_submodules(model, key) -> Tuple:
    """
    Retrieves the parent module, target module, and target module name for a given key in a PyTorch model.
    """
    parent = model.get_submodule(".".join(key.split(".")[:-1]))
    target_name = key.split(".")[-1]
    target = model.get_submodule(key)
    return parent, target, target_name


def linearize_lora_model_(model):
    """
    Linearizes the LoraLayer modules in a PyTorch model according to the PETA paper.
    """
    for key, module in model.named_modules():
        # if isinstance(module, LoraLayer) and isinstance(module, nn.Linear):
        if isinstance(module, LoraLayer):
            # print("L-LoRA MODULE : ", module)
            parent, target, target_name = _get_submodules(model, key)
            setattr(parent, target_name, LinearizedModelWraper(target))
            # print("Linearized Lora Layer")
    return model


def unlinearize_lora_model_(model):
    """
    Unloads the linearized LoraLayer modules in a PyTorch model.
    """
    LinearizedModelWraper.unload_linearized_modules_(model)
    return model


class ImageClassificationFineTuningForCLIP(
    CLIPClassificationMixin,
    SimpleProfilerMixin,
    ModelFusionAlgorithm,
):
    """
    A class for fine-tuning CLIP models for image classification tasks.
    """

    def run(self, modelpool: CLIPVisionModelPool):
        """
        Executes the fine-tuning process.

        Args:
            modelpool (CLIPVisionModelPool): The modelpool is responsible for loading the pre-trained model and training datasets.

        Returns:
            VisionModel: The fine-tuned vision model.
        """
        self.modelpool = to_modelpool(modelpool)
        config = self.config
        self.log_hyperparams(config, filename="method_config.yaml")
        self.finetune_method = "fine-tune"

        L.seed_everything(config.seed)

        task_names = modelpool.train_dataset_names
        with self.profile("setup model and optimizer"):
            processor, classifier, optimizer, lr_scheduler = self.setup_model()

            if config.state_dict_load_path is not None:
                self.fabric.load(
                    config.state_dict_load_path,
                    {"vision_model": classifier.clip_model.vision_model},
                )
                if config.skip_training:
                    return classifier.clip_model.vision_model

            self.setup_zero_shot_classification_head(
                clip_processor=processor,
                clip_model=classifier.clip_model,
                task_names=task_names,
            )

            self.fabric.setup(classifier, optimizer)

        with self.profile("setup data"):
            train_datasets = [
                CLIPDataset(modelpool.load_train_dataset(task_name), processor)
                for task_name in task_names
            ]
            train_dataloaders = [
                DataLoader(
                    dataset,
                    shuffle=True,
                    batch_size=config.batch_size,
                    num_workers=config.num_workers,
                )
                for dataset in train_datasets
            ]
            train_dataloaders = self.fabric.setup_dataloaders(*train_dataloaders)
            if not isinstance(train_dataloaders, (list, tuple)):
                train_dataloaders = [train_dataloaders]
            train_dataloader_iters = [
                iter(InfiniteDataLoader(loader)) for loader in train_dataloaders
            ]

        # train
        for step_idx in tqdm(
            range(config.num_steps),
            desc=self.finetune_method,
            disable=not self.fabric.is_global_zero,
            dynamic_ncols=True,
        ):
            optimizer.zero_grad()
            loss = 0
            for task, loader in zip(task_names, train_dataloader_iters):
                with self.profile("data loading"):
                    batch = next(loader)
                    images, labels = batch
                with self.profile("forward"):
                    classifier.zeroshot_weights = self.zeroshot_weights[task]
                    logits = classifier(images)
                loss = loss + nn.functional.cross_entropy(logits, labels)

            with self.profile("backward"):
                self.fabric.backward(loss)
            with self.profile("optimizer step"):
                optimizer.step()
                lr_scheduler.step()

            metrics = {"train/loss": loss}

            self.fabric.log_dict(metrics, step=step_idx)

            if (step_idx + 1) % config.save_interval == 0:
                save_path = os.path.join(
                    self.log_dir, "checkpoints", f"step={step_idx}.ckpt"
                )
                self.save_model(classifier, save_path)

        if config.state_dict_save_path is not None:
            self.save_model(classifier, config.state_dict_save_path)
        self.print_profile_summary()
        return classifier.clip_model.vision_model

    def save_model(
        self,
        model: HFCLIPClassifier | CLIPModel | CLIPVisionModel | CLIPVisionTransformer,
        save_path: str,
    ):
        """
        Save the vision model to the specified path.

        Args:
            model (Union[HFCLIPClassifier, CLIPModel, CLIPVisionModel, CLIPVisionTransformer]): The model to save.
            save_path (str): The path to save the model.
        """
        if isinstance(model, HFCLIPClassifier):
            vision_model = model.clip_model.vision_model
        elif isinstance(model, CLIPModel):
            vision_model = model.vision_model
        elif isinstance(model, CLIPVisionModel):
            vision_model = model.vision_model
        elif isinstance(model, CLIPVisionTransformer):
            vision_model = model
        else:
            raise ValueError(f"Unsupported model type: {type(model)}")

        save_dir = os.path.dirname(save_path)
        if save_dir and not os.path.exists(save_dir):
            os.makedirs(save_dir, exist_ok=True)
        self.fabric.save(save_path, {"vision_model": vision_model})

    def setup_model(self):
        """
        Sets up the model, optimizer, and learning rate scheduler.

        This method initializes the CLIP model, applies LoRA if specified, and configures the optimizer and learning rate scheduler.

        Returns:
            Tuple: A tuple containing the processor, classifier, optimizer, and learning rate scheduler.
        """
        config = self.config
        modelpool = self.modelpool

        clip_model: CLIPModel = modelpool.load_clip_model("_pretrained_")
        processor = modelpool.load_processor()

        self.finetune_method = "full fine-tune"
        if config.use_lora or config.use_l_lora:
            self.finetune_method = "lora fine-tune"
            lora_config = LoraConfig(
                **OmegaConf.to_container(
                    config.lora_config, resolve=True, enum_to_str=True
                )
            )
            clip_model.vision_model = get_peft_model(
                clip_model.vision_model, lora_config
            )

            if config.use_l_lora:
                # http://arxiv.org/abs/2310.04742
                # Anke Tang et al. Parameter Efficient Multi-task Model Fusion with Partial Linearization. ICLR 2024.
                self.finetune_method = "l-lora fine-tune"
                print("Linearizing Lora Layers")
                linearize_lora_model_(clip_model.vision_model)

        classifier = HFCLIPClassifier(clip_model, processor=processor)

        if self.fabric.is_global_zero:
            print("=== Model Summary (For Vision Model Only) ===")
            print_parameters(classifier.clip_model.vision_model)
        # configure optimizers
        optimizer = torch.optim.Adam(
            [
                p
                for p in classifier.clip_model.vision_model.parameters()
                if p.requires_grad
            ],
            lr=config.learning_rate,
            weight_decay=config.weight_decay,
        )
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer=optimizer, T_max=config.num_steps
        )

        return processor, classifier, optimizer, lr_scheduler


def load_full_finetuned_vision_model(
    pretrained_path: str, state_dict_path: str, strict=True
):
    """
    Load a fully fine-tuned model from the state_dict_path and pretrained_path.

    Args:
        pretrained_path (str): The path to the pretrained model.
        state_dict_path (str): The path to the state dictionary.
        strict (bool, optional): Whether to strictly enforce that the keys in state_dict match the keys returned by this module's state_dict() function. Defaults to True.

    Returns:
        CLIPVisionModel: The loaded vision model.
    """
    model: CLIPVisionModel = CLIPVisionModel.from_pretrained(pretrained_path)
    model.vision_model.load_state_dict(
        torch.load(state_dict_path, map_location="cpu")["vision_model"], strict=strict
    )
    return model


def load_lora_vision_model(
    pretrained_path: str,
    lora_config: LoraConfig,
    state_dict_path: str,
) -> PeftModel:
    """
    Load LoRA model from the state_dict_path and pretrained_path.

    Args:
        pretrained_path (str): The path to the pretrained model.
        lora_config (LoraConfig): The configuration for LoRA.
        state_dict_path (str): The path to the state dictionary.

    Returns:
        PeftModel: The loaded LoRA model.
    """
    model: CLIPVisionModel = CLIPVisionModel.from_pretrained(pretrained_path)
    model = get_peft_model(model.vision_model, lora_config)
    state_dict = torch.load(state_dict_path, map_location="cpu")["vision_model"]
    for name, value in state_dict.items():
        model.get_parameter(name).data = value
    return model


def load_l_lora_vision_model(
    pretrained_path: str,
    lora_config: LoraConfig,
    state_dict_path: str,
    unload_linearized_modules: bool = False,
):
    """
    Load L-LoRA model from the state_dict_path and pretrained_path.

    The output folder should contain the following files:

    - README.md
    - adapter_config.json
    - linearized_adapter_model.safetensors

    Load the converted model using the following code:

    >>> from fusion_bench.models.linearized.vision_model import load_l_lora_vision_model_hf
    >>> model = load_l_lora_vision_model_hf("base_model_name", "peft_name")

    Args:
        pretrained_path (str): The path to the pretrained model.
        lora_config (LoraConfig): The configuration for LoRA.
        state_dict_path (str): The path to the state dictionary.
        unload_linearized_modules (bool, optional): Whether to unload the linearized modules. Defaults to False.

    Returns:
        PeftModel: The loaded L-LoRA model.
    """
    model: CLIPVisionModel = CLIPVisionModel.from_pretrained(pretrained_path)
    model = get_peft_model(model.vision_model, lora_config)
    linearize_lora_model_(model)
    state_dict = torch.load(state_dict_path, map_location="cpu")["vision_model"]
    for name, value in state_dict.items():
        model.get_parameter(name).data = value
    if unload_linearized_modules:
        LinearizedModelWraper.unload_linearized_modules_(model)
    return model


def convert_lora_state_dict_to_hf(
    pretrained_path: str,
    ckpt_path: str,
    lora_config: LoraConfig,
    output_path: str,
    base_model_name: Optional[str] = None,
):
    """
    Convert a LoRA model's checkpoint to Huggingface's format.

    Args:
        pretrained_path (str): The path to the pretrained model.
        ckpt_path (str): The path to the checkpoint.
        lora_config (LoraConfig): The configuration for LoRA.
        output_path (str): The path to save the converted model.
        base_model_name (Optional[str], optional): The name of the base model. Defaults to None.
    """
    model = load_lora_vision_model(
        pretrained_path=pretrained_path,
        state_dict_path=ckpt_path,
        lora_config=lora_config,
    )

    model.config._name_or_path = base_model_name
    model.peft_config["default"].base_model_name_or_path = base_model_name
    model.save_pretrained(output_path)


def convert_l_lora_state_dict_to_hf(
    pretrained_path: str,
    ckpt_path: str,
    lora_config: LoraConfig,
    output_path: str,
    base_model_name: Optional[str] = None,
):
    """
    Convert a linearized Lora model's checkpoint to huggingface's format.

    Args:
        pretrained_path (str): The path to the pretrained model.
        ckpt_path (str): The path to the checkpoint.
        lora_config (LoraConfig): The configuration for LoRA.
        output_path (str): The path to save the converted model.
        base_model_name (Optional[str], optional): The name of the base model. Defaults to None.
    """

    if base_model_name is None:
        base_model_name = pretrained_path

    # load l-lora model from ckpt file
    model = load_l_lora_vision_model(
        pretrained_path=pretrained_path,
        state_dict_path=ckpt_path,
        lora_config=lora_config,
        unload_linearized_modules=False,
    )

    # save lora config
    model.peft_config["default"].base_model_name_or_path = base_model_name
    model.peft_config["default"].save_pretrained(output_path)

    # save linearized adapter to safetensors
    linearized_adapter = {}
    for name, module in model.named_modules():
        if isinstance(module, LinearizedModelWraper):
            for (param_name, param), (param0_name, param0) in zip(
                module.model.named_parameters(),
                module.params0_values.named_parameters(),
            ):
                if param.requires_grad:
                    linearized_adapter[f"{name}.model.{param_name}"] = param
                    linearized_adapter[f"{name}.params0_values.{param0_name}"] = param0

    save_file(linearized_adapter, f"{output_path}/linearized_adapter_model.safetensors")

    readme = f"""---
base_model: {base_model_name}
library_name: peft
---

# Model Card L-LoRA Model
"""
    # open README.md
    with open(f"{output_path}/README.md", "w") as file:
        file.write(readme)



"""
Online documentation for this module: https://tanganke.github.io/fusion_bench/modelpool/causal_lm
"""

import logging
import os
from copy import deepcopy
from typing import Any, Dict, Optional, TypeAlias, Union, cast  # noqa: F401

import peft
from omegaconf import DictConfig, OmegaConf, flag_override
from torch import nn
from torch.nn.modules import Module
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from typing_extensions import override

from fusion_bench import (
    BaseModelPool,
    auto_register_config,
    import_object,
    instantiate,
    parse_dtype,
)
from fusion_bench.utils.lazy_state_dict import LazyStateDict

log = logging.getLogger(__name__)


@auto_register_config
class CausalLMPool(BaseModelPool):
    """A model pool for managing and loading causal language models.

    This class provides a unified interface for loading and managing multiple
    causal language models, typically used in model fusion and ensemble scenarios.
    It supports both eager and lazy loading strategies, and handles model
    configuration through YAML configs or direct instantiation.

    The pool can manage models from Hugging Face Hub, local paths, or custom
    configurations. It also provides tokenizer management and model saving
    capabilities with optional Hugging Face Hub integration.

    Args:
        models: Dictionary or configuration specifying the models to be managed.
            Can contain model names mapped to paths or detailed configurations.
        tokenizer: Tokenizer configuration, either a string path/name or
            a DictConfig with detailed tokenizer settings.
        model_kwargs: Additional keyword arguments passed to model loading.
            Common options include torch_dtype, device_map, etc.
        enable_lazy_loading: Whether to use lazy loading for models. When True,
            models are loaded as LazyStateDict objects instead of actual models,
            which can save memory for large model collections.
        **kwargs: Additional arguments passed to the parent BaseModelPool.

    Example:
        ```python
        >>> pool = CausalLMPool(
        ...     models={
        ...         "model_a": "microsoft/DialoGPT-medium",
        ...         "model_b": "/path/to/local/model"
        ...     },
        ...     tokenizer="microsoft/DialoGPT-medium",
        ...     model_kwargs={"torch_dtype": "bfloat16"}
        ... )
        >>> model = pool.load_model("model_a")
        >>> tokenizer = pool.load_tokenizer()
        ```
    """

    def __init__(
        self,
        models,
        *,
        tokenizer: Optional[DictConfig | str],
        model_kwargs: Optional[DictConfig] = None,
        enable_lazy_loading: bool = False,
        **kwargs,
    ):
        super().__init__(models, **kwargs)
        if model_kwargs is None:
            self.model_kwargs = DictConfig({})

    def get_model_path(self, model_name: str):
        """Extract the model path from the model configuration.

        Args:
            model_name: The name of the model as defined in the models configuration.

        Returns:
            str: The path or identifier for the model. For string configurations,
                returns the string directly. For dict configurations, extracts
                the 'pretrained_model_name_or_path' field.

        Raises:
            RuntimeError: If the model configuration is invalid or the model
                name is not found in the configuration.
        """
        model_name_or_config = self._models[model_name]
        if isinstance(model_name_or_config, str):
            return model_name_or_config
        elif isinstance(model_name_or_config, (DictConfig, dict)):
            return model_name_or_config.get("pretrained_model_name_or_path")
        else:
            raise RuntimeError("Invalid model configuration")

    def get_model_kwargs(self):
        """Get processed model keyword arguments for model loading.

        Converts the stored `model_kwargs` from DictConfig to a regular dictionary
        and processes special arguments like torch_dtype for proper model loading.

        Returns:
            dict: Processed keyword arguments ready to be passed to model
                loading functions. The torch_dtype field, if present, is
                converted from string to the appropriate torch dtype object.
        """
        model_kwargs = (
            OmegaConf.to_container(self.model_kwargs, resolve=True)
            if isinstance(self.model_kwargs, DictConfig)
            else self.model_kwargs
        )
        if "torch_dtype" in model_kwargs:
            model_kwargs["torch_dtype"] = parse_dtype(model_kwargs["torch_dtype"])
        return model_kwargs

    @override
    def load_model(
        self,
        model_name_or_config: str | DictConfig,
        *args,
        **kwargs,
    ) -> Union[PreTrainedModel, LazyStateDict]:
        """Load a causal language model from the model pool.

        This method supports multiple loading strategies:
        1. Loading by model name from the configured model pool
        2. Loading from a direct configuration dictionary
        3. Lazy loading using LazyStateDict for memory efficiency

        The method automatically handles different model configuration formats
        and applies the appropriate loading strategy based on the enable_lazy_loading flag.

        Args:
            model_name_or_config: Either a string model name that exists in the
                model pool configuration, or a DictConfig/dict containing the
                model configuration directly.
            *args: Additional positional arguments passed to the model constructor.
            **kwargs: Additional keyword arguments passed to the model constructor.
                These will be merged with the pool's model_kwargs.

        Returns:
            Union[PreTrainedModel, LazyStateDict]: The loaded model. Returns a
                PreTrainedModel for normal loading or a LazyStateDict for lazy loading.

        Raises:
            RuntimeError: If the model configuration is invalid.
            KeyError: If the model name is not found in the model pool.

        Example YAML configurations:
            Simple string configuration:
            ```yaml
            models:
              _pretrained_: path_to_pretrained_model
              model_a: path_to_model_a
              model_b: path_to_model_b
            ```

            Detailed configuration:
            ```yaml
            models:
              _pretrained_:
                _target_: transformers.AutoModelForCausalLM
                pretrained_model_name_or_path: path_to_pretrained_model
              model_a:
                _target_: transformers.AutoModelForCausalLM
                pretrained_model_name_or_path: path_to_model_a
            ```
        """
        model_kwargs = self.get_model_kwargs()
        model_kwargs.update(kwargs)

        if isinstance(model_name_or_config, str):
            # If model_name_or_config is a string, it is the name or the path of the model
            log.info(f"Loading model: {model_name_or_config}", stacklevel=2)
            if model_name_or_config in self._models.keys():
                model_config = self._models[model_name_or_config]
                if isinstance(model_config, str):
                    # model_config is a string
                    if not self.enable_lazy_loading:
                        model = AutoModelForCausalLM.from_pretrained(
                            model_config,
                            *args,
                            **model_kwargs,
                        )
                    else:
                        # model_config is a string, but we want to use LazyStateDict
                        model = LazyStateDict(
                            checkpoint=model_config,
                            meta_module_class=AutoModelForCausalLM,
                            *args,
                            **model_kwargs,
                        )
                    return model
        elif isinstance(model_name_or_config, (DictConfig, Dict)):
            model_config = model_name_or_config

        if not self.enable_lazy_loading:
            model = instantiate(model_config, *args, **model_kwargs)
        else:
            meta_module_class = model_config.pop("_target_")
            checkpoint = model_config.pop("pretrained_model_name_or_path")
            model = LazyStateDict(
                checkpoint=checkpoint,
                meta_module_class=meta_module_class,
                *args,
                **model_kwargs,
            )
        return model

    def load_tokenizer(self, *args, **kwargs) -> PreTrainedTokenizer:
        """Load the tokenizer associated with this model pool.

        Loads a tokenizer based on the tokenizer configuration provided during
        pool initialization. Supports both simple string paths and detailed
        configuration dictionaries.

        Args:
            *args: Additional positional arguments passed to the tokenizer constructor.
            **kwargs: Additional keyword arguments passed to the tokenizer constructor.

        Returns:
            PreTrainedTokenizer: The loaded tokenizer instance.

        Raises:
            AssertionError: If no tokenizer is defined in the configuration.

        Example YAML configurations:
            Simple string configuration:
            ```yaml
            tokenizer: google/gemma-2-2b-it
            ```

            Detailed configuration:
            ```yaml
            tokenizer:
              _target_: transformers.AutoTokenizer
              pretrained_model_name_or_path: google/gemma-2-2b-it
              use_fast: true
              padding_side: left
            ```
        """
        assert self.tokenizer is not None, "Tokenizer is not defined in the config"
        log.info("Loading tokenizer.", stacklevel=2)
        if isinstance(self.tokenizer, str):
            tokenizer = AutoTokenizer.from_pretrained(self.tokenizer, *args, **kwargs)
        else:
            tokenizer = instantiate(self.tokenizer, *args, **kwargs)
        return tokenizer

    @override
    def save_model(
        self,
        model: PreTrainedModel,
        path: str,
        push_to_hub: bool = False,
        model_dtype: Optional[str] = None,
        save_tokenizer: bool = False,
        tokenizer_kwargs=None,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        **kwargs,
    ):
        """Save a model to the specified path with optional tokenizer and Hub upload.

        This method provides comprehensive model saving capabilities including
        optional tokenizer saving, dtype conversion, and Hugging Face Hub upload.
        The model is saved in the standard Hugging Face format.

        Args:
            model: The PreTrainedModel instance to be saved.
            path: The local path where the model will be saved. Supports tilde
                expansion for home directory paths.
            push_to_hub: Whether to push the saved model to the Hugging Face Hub.
                Requires proper authentication and repository permissions.
            model_dtype: Optional string specifying the target dtype for the model
                before saving (e.g., "float16", "bfloat16"). The model will be
                converted to this dtype before saving.
            save_tokenizer: Whether to save the tokenizer alongside the model.
                If True, the tokenizer will be loaded using the pool's tokenizer
                configuration and saved to the same path.
            tokenizer_kwargs: Additional keyword arguments for tokenizer loading
                when save_tokenizer is True.
            tokenizer: Optional pre-loaded tokenizer instance. If provided, this
                tokenizer will be saved regardless of the save_tokenizer flag.
            **kwargs: Additional keyword arguments passed to the model's
                save_pretrained method.

        Side Effects:
            - Creates model files in the specified directory
            - Optionally creates tokenizer files in the same directory
            - May convert the model to a different dtype
            - May upload files to Hugging Face Hub

        Example:
            ```python
            >>> pool = CausalLMPool(models=..., tokenizer=...)
            >>> model = pool.load_model("my_model")
            >>> pool.save_model(
            ...     model,
            ...     "/path/to/save",
            ...     save_tokenizer=True,
            ...     model_dtype="float16",
            ...     push_to_hub=True
            ... )
            ```
        """
        path = os.path.expanduser(path)
        # NOTE: if tokenizer is provided, it will be saved regardless of `save_tokenizer`
        if save_tokenizer or tokenizer is not None:
            if tokenizer is None:
                if tokenizer_kwargs is None:
                    tokenizer_kwargs = {}
                # load the tokenizer
                tokenizer = self.load_tokenizer(**tokenizer_kwargs)
            tokenizer.save_pretrained(
                path,
                push_to_hub=push_to_hub,
            )
        if model_dtype is not None:
            model.to(dtype=parse_dtype(model_dtype))
        model.save_pretrained(
            path,
            push_to_hub=push_to_hub,
            **kwargs,
        )


class CausalLMBackbonePool(CausalLMPool):
    """A specialized model pool that loads only the transformer backbone layers.

    This class extends CausalLMPool to provide access to just the transformer
    layers (backbone) of causal language models, excluding the language modeling
    head and embeddings. This is useful for model fusion scenarios where only
    the core transformer layers are needed.

    The class automatically extracts the `model.layers` component from loaded
    AutoModelForCausalLM instances, providing direct access to the transformer
    blocks. Lazy loading is not supported for this pool type.

    Note:
        This pool automatically disables lazy loading as it needs to access
        the internal structure of the model to extract the backbone layers.

    Example:
        ```python
        >>> backbone_pool = CausalLMBackbonePool(
        ...     models={"model_a": "microsoft/DialoGPT-medium"},
        ...     tokenizer="microsoft/DialoGPT-medium"
        ... )
        >>> layers = backbone_pool.load_model("model_a")  # Returns nn.ModuleList of transformer layers
        ```
    """

    def load_model(
        self, model_name_or_config: str | DictConfig, *args, **kwargs
    ) -> Module:
        """Load only the transformer backbone layers from a causal language model.

        This method loads a complete causal language model and then extracts
        only the transformer layers (backbone), discarding the embedding layers
        and language modeling head. This is useful for model fusion scenarios
        where only the core transformer computation is needed.

        Args:
            model_name_or_config: Either a string model name from the pool
                configuration or a DictConfig with model loading parameters.
            *args: Additional positional arguments passed to the parent load_model method.
            **kwargs: Additional keyword arguments passed to the parent load_model method.

        Returns:
            Module: The transformer layers (typically a nn.ModuleList) containing
                the core transformer blocks without embeddings or output heads.

        Note:
            Lazy loading is automatically disabled for this method as it needs
            to access the internal model structure to extract the layers.
        """
        if self.enable_lazy_loading:
            log.warning(
                "CausalLMBackbonePool does not support lazy loading. "
                "Falling back to normal loading."
            )
            self.enable_lazy_loading = False
        model: AutoModelForCausalLM = super().load_model(
            model_name_or_config, *args, **kwargs
        )
        return model.model.layers


def load_peft_causal_lm(
    base_model_path: str,
    peft_model_path: str,
    torch_dtype: str = "bfloat16",
    is_trainable: bool = True,
    merge_and_unload: bool = False,
):
    """Load a causal language model with PEFT (Parameter-Efficient Fine-Tuning) adapters.

    This function loads a base causal language model and applies PEFT adapters
    (such as LoRA, AdaLoRA, or other parameter-efficient fine-tuning methods)
    to create a fine-tuned model. It supports both keeping the adapters separate
    or merging them into the base model.

    Args:
        base_model_path: Path or identifier for the base causal language model.
            Can be a Hugging Face model name or local path.
        peft_model_path: Path to the PEFT adapter configuration and weights.
            This should contain the adapter_config.json and adapter weights.
        torch_dtype: The torch data type to use for the model. Common options
            include "float16", "bfloat16", "float32". Defaults to "bfloat16".
        is_trainable: Whether the loaded PEFT model should be trainable.
            Set to False for inference-only usage to save memory.
        merge_and_unload: Whether to merge the PEFT adapters into the base model
            and unload the adapter weights. When True, returns a standard
            PreTrainedModel instead of a PeftModel.

    Returns:
        Union[PeftModel, PreTrainedModel]: The loaded model with PEFT adapters.
            Returns a PeftModel if merge_and_unload is False, or a PreTrainedModel
            if the adapters are merged and unloaded.

    Example:
        ```python
        >>> # Load model with adapters for training
        >>> model = load_peft_causal_lm(
        ...     "microsoft/DialoGPT-medium",
        ...     "/path/to/lora/adapters",
        ...     is_trainable=True
        ... )

        >>> # Load and merge adapters for inference
        >>> merged_model = load_peft_causal_lm(
        ...     "microsoft/DialoGPT-medium",
        ...     "/path/to/lora/adapters",
        ...     merge_and_unload=True,
        ...     is_trainable=False
        ... )
        ```
    """
    base_model = AutoModelForCausalLM.from_pretrained(
        base_model_path, torch_dtype=torch_dtype
    )
    model = peft.PeftModel.from_pretrained(
        base_model,
        peft_model_path,
        is_trainable=is_trainable,
    )
    if merge_and_unload:
        model = model.merge_and_unload()
    return model



import copy
import functools
import logging
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Dict, Iterator, List, Optional  # noqa: F401

import lightning as L
import torch
from torch import Tensor, nn
from torch.func import functional_call

from fusion_bench.models.utils import del_attr, get_attr, set_attr
from fusion_bench.utils.state_dict_arithmetic import state_dict_add
from fusion_bench.utils.type import StateDictType

from .layer_wise_fusion import fuse_weights, get_layer_wise_weights

__all__ = ["get_layer_wise_weights", "fuse_weights", "LayerWiseMergedModel"]

log = logging.getLogger(__name__)


class LayerWiseMergedModel(nn.Module):
    _merged_state_dict: StateDictType = None

    def __init__(
        self,
        layer_wise_weight: Tensor,
        pretrained_model: nn.Module,
        finetuned_models: List[nn.Module],
        clamp_weights: bool = True,
        tie_weights: bool = False,
        strict: bool = True,
        sparsity_ratio: Optional[float] = None,
        normalized_merging_weights: bool = False,
    ):
        R"""
        This class wraps a pretrained model and a list of finetuned models, and merges the weights of the finetuned models into the pretrained model using layer-wise fusion.

        Args:
            layer_wise_weight (Tensor): A tensor of shape (num_models, num_layers) representing the weight of each layer for each model.
            pretrained_model (nn.Module): The pretrained model to merge the weights into.
            finetuned_models (List[nn.Module]): A list of finetuned models to merge the weights from. This should have the same architecture as the pretrained model. We use these models to compute the task vectors.
            clamp_weights (bool, optional): If True, the layer-wise weights will be clamped to [0, 1]. Defaults to True.
            tie_weights (bool, optional): This option passes the `tie_weights` argument to the `functional_call` function. Defaults to False.
            strict (bool, optional): This option passes the `strict` argument to the `functional_call` function. Defaults to True.
            sparsity_ratio (float, optional): If `sparsity_ratio` is provided, the task vector will be pruned before merging. A high spasity level can save the memory usage during merging.
            normalized_merging_weights (bool, optional): If True, the layer-wise weights will be normalized for each layer, so that the sum of weights across models for each layer is 1. Defaults to False.
        """
        super().__init__()
        if torch.cuda.is_available():
            self._fabric = L.Fabric(devices=1)
            self._fabric.launch()
        self.clamp_weights = clamp_weights
        self.tie_weights = tie_weights
        self.strict = strict
        self.sparsity_ratio = sparsity_ratio
        self.nromalized_merging_weights = normalized_merging_weights

        pretrained_sd = pretrained_model.state_dict(keep_vars=True)
        filtered_keys = [
            k
            for k in pretrained_sd.keys()
            if ("encoder" in k and "layer_norm" not in k and "weight" in k)
        ]
        self.merge_weight = nn.Parameter(
            layer_wise_weight[:, : len(filtered_keys)], requires_grad=True
        )
        task_vectors = []
        for m in finetuned_models:
            m.requires_grad_(False)
        self.pretrained_model = pretrained_model.requires_grad_(False)
        for model in finetuned_models:
            model_sd = model.state_dict(keep_vars=True)
            filtered_task_vector = {
                k: (model_sd[k] - pretrained_sd[k]) for k in filtered_keys
            }
            if self._fabric is not None:
                filtered_task_vector = self._fabric.to_device(filtered_task_vector)
            task_vectors.append(filtered_task_vector)

        self.projection = {}
        for layer_name in task_vectors[0].keys():
            for i, vector in enumerate(task_vectors):
                layer_vector = vector[layer_name]
                u, s, v = torch.linalg.svd(layer_vector, full_matrices=False)
                if i == 0:
                    print(f"Computed SVD for {layer_name}...")
                    sum_u = torch.zeros_like(u, device=layer_vector.device)
                    sum_s = torch.zeros_like(s, device=layer_vector.device)
                    sum_v = torch.zeros_like(v, device=layer_vector.device)

                reduced_index_s = int(s.shape[0] / len(task_vectors))

                # select only the first reduced_index_s columns of u and place them
                sum_u[:, i * reduced_index_s : (i + 1) * reduced_index_s] = u[
                    :, :reduced_index_s
                ]
                sum_s[i * reduced_index_s : (i + 1) * reduced_index_s] = s[
                    :reduced_index_s
                ]
                # select only the first reduced_index_s rows of v and place them
                sum_v[i * reduced_index_s : (i + 1) * reduced_index_s, :] = v[
                    :reduced_index_s, :
                ]
            u_u, s_u, v_u = torch.linalg.svd(sum_u, full_matrices=False)
            # u_v, s_v, v_v = torch.linalg.svd(sum_v, full_matrices=False)
            layer_proj = torch.matmul(
                u_u[:, : int(s.shape[0] / len(task_vectors))],
                u_u[:, : int(s.shape[0] / len(task_vectors))].T,
            )
            self.projection[layer_name] = layer_proj

        self.delta = [
            {
                k: torch.zeros_like(v).clone().requires_grad_()
                for k, v in task_vector.items()
            }
            for task_vector in task_vectors
        ]
        if self._fabric is not None:
            self.delta = self._fabric.to_device(self.delta)
        self.lamdas = self.compute_layer_lamdas(task_vectors)

        for layer_name in task_vectors[0].keys():
            optimizer = torch.optim.Adam(
                [delta[layer_name] for delta in self.delta], lr=1e-4
            )
            layer_vectors = torch.stack([vec[layer_name] for vec in task_vectors])
            layer_lamdas = torch.stack([lamdas[layer_name] for lamdas in self.lamdas])
            for _ in range(400):
                optimizer.zero_grad()
                layer_delta = torch.stack([de[layer_name] for de in self.delta])
                loss = self.taskvector_loss(layer_vectors, layer_delta, layer_lamdas)
                print(f"Epoch: {_}, Layer: {layer_name}, Loss: {loss.item()}")
                self._fabric.backward(loss)
                for delta in self.delta:
                    grad_proj = (
                        self.projection[layer_name] @ delta[layer_name].grad.detach()
                    )
                    delta[layer_name].grad.data = delta[layer_name].grad.data.sub_(
                        grad_proj
                    )
                optimizer.step()
                for delta in self.delta:
                    for param in delta.values():
                        param.grad = None
        del self.projection
        self.delta = [
            {key: param.detach().cpu() for key, param in delta.items()}
            for delta in self.delta
        ]
        self.lamdas = [
            {key: param.cpu() for key, param in lamdas.items()}
            for lamdas in self.lamdas
        ]
        task_vectors = [
            {k: v.cpu() for k, v in task_vector.items()} for task_vector in task_vectors
        ]
        flat_vectors = []
        vector_masks = []
        for idx, task_vector in enumerate(task_vectors):
            flat_vector = self.state_dict_to_vector(task_vector)
            vector_mask = self.topk_values_mask(flat_vector, K=30)
            flat_vectors.append(flat_vector)
            vector_masks.append(vector_mask)
        flat_deltas = [self.state_dict_to_vector(delta) for delta in self.delta]
        self.task_vectors = [
            self.vector_to_state_dict(
                (flat_vector + flat_delta) * vector_mask, self.delta[0]
            )
            for flat_vector, flat_delta, vector_mask in zip(
                flat_vectors, flat_deltas, vector_masks
            )
        ]
        if self._fabric is not None:
            self.task_vectors = self._fabric.to_device(self.task_vectors)

        # if `sparisty_ratio` is given, pruning the task vectors.
        if sparsity_ratio is not None:
            from fusion_bench.method.pruning.prune_utils import (
                unstructured_magnitude_prune_,
            )

            for name, param in self.task_vectors.named_parameters():
                if param.dim() != 2:
                    continue
                print(f"pruning {name}")
                pruned_param = unstructured_magnitude_prune_(
                    param.data.clone(), torch.abs, sparsity_ratio=sparsity_ratio
                )
                set_attr(
                    self.task_vectors,
                    name.split("."),
                    nn.Parameter(pruned_param.to_sparse(), requires_grad=False),
                )

    def topk_values_mask(self, M, K):
        if K > 1:
            K /= 100

        original_shape = M.shape
        if M.dim() == 1:
            M = M.unsqueeze(0)

        n, d = M.shape
        k = int(d * K)
        k = d - k  # Keep top k elements instead of bottom k elements

        # Find the k-th smallest element by magnitude for each row
        kth_values, _ = M.abs().kthvalue(k, dim=1, keepdim=True)
        # Create a mask tensor with True for the top k elements in each row
        mask = M.abs() >= kth_values
        final_mask = mask.squeeze() if original_shape == M.squeeze().shape else mask

        return final_mask

    def state_dict_to_vector(self, state_dict, remove_keys=[]):
        """
        Convert a state dictionary to a vector, removing specified keys.

        Args:
            state_dict (dict): The state dictionary to convert.
            remove_keys (list): List of keys to remove from the state dictionary.

        Returns:
            Tensor: A vector representation of the state dictionary.
        """
        shared_state_dict = copy.deepcopy(state_dict)
        for key in remove_keys:
            if key in shared_state_dict:
                del shared_state_dict[key]
        sorted_shared_state_dict = OrderedDict(sorted(shared_state_dict.items()))
        return nn.utils.parameters_to_vector(
            [value.reshape(-1) for key, value in sorted_shared_state_dict.items()]
        )

    def vector_to_state_dict(self, vector, state_dict, remove_keys=[]):
        """
        Convert a vector back to a state dictionary, removing specified keys.

        Args:
            vector (Tensor): The vector to convert.
            state_dict (dict): The reference state dictionary.
            remove_keys (list): List of keys to remove from the state dictionary.

        Returns:
            dict: A state dictionary representation of the vector.
        """
        # create a reference dict to define the order of the vector
        reference_dict = copy.deepcopy(state_dict)
        for key in remove_keys:
            if key in reference_dict:
                del reference_dict[key]
        sorted_reference_dict = OrderedDict(sorted(reference_dict.items()))

        # create a shared state dict using the reference dict
        nn.utils.vector_to_parameters(vector, sorted_reference_dict.values())

        # add back the encoder and decoder embedding weights.
        if "transformer.shared.weight" in sorted_reference_dict:
            for key in remove_keys:
                sorted_reference_dict[key] = sorted_reference_dict[
                    "transformer.shared.weight"
                ]
        return sorted_reference_dict

    def taskvector_loss(self, layer_vectors, layer_delta, layer_lamdas) -> torch.Tensor:
        """
        Computes the loss based on delta and task vectors for a specific layer.
        """
        total_loss = 0.0

        layer_vectors_scale = layer_vectors * layer_lamdas.view(-1, 1, 1)
        sum_over_num_vectors = layer_vectors_scale.sum(dim=0)

        layer_delta_scale = layer_delta * layer_lamdas.view(-1, 1, 1)
        sum_over_delta = layer_delta_scale.sum(dim=0)

        # Iterate through each vector and calculate the loss one by one
        for v_j in layer_vectors:
            part1 = -v_j * sum_over_num_vectors
            part2 = -v_j * sum_over_delta
            part3 = v_j * v_j

            expression = part1 + part2 + part3
            layer_loss = expression.sum(dim=1).pow(2).sum()

            # Cumulative total loss
            total_loss += layer_loss
        return total_loss

    def compute_layer_lamdas(self, vectors: List[StateDictType]) -> torch.Tensor:
        lamdas = []
        for vec in vectors:
            tmp = {}
            for layer_name in vec.keys():
                norm_vec = torch.norm(vec[layer_name])
                tmp[layer_name] = 0.07 / norm_vec
            lamdas.append(tmp)
        return lamdas

    @property
    def forward_model(self):
        return functools.partial(
            functional_call,
            self.pretrained_model,
            self._merged_state_dict,
            tie_weights=self.tie_weights,
            strict=self.strict,
        )

    def merge_and_unload(self, task_vector_mask: Optional[Dict[str, Tensor]] = None):
        self.merge_weights(task_vector_mask=task_vector_mask)
        self.pretrained_model.load_state_dict(self._merged_state_dict)
        return self.pretrained_model

    def merge_weights(self, task_vector_mask: Optional[Dict[str, Tensor]] = None):
        """
        Merges the weights of the model.
        Call this after each update step.
        """
        if self.clamp_weights:
            layer_wise_weight = self.merge_weight.clamp(0, 1)
        else:
            layer_wise_weight = self.merge_weight
        if self.nromalized_merging_weights:
            # normalize the weights for each layer, so that the sum of weights across models for each layer is 1.
            layer_wise_weight = layer_wise_weight.softmax(dim=0)

        state_dict = self.pretrained_model.state_dict(keep_vars=True)
        # shape of layer_wise_weight: (num_models, num_layers)
        for weight, task_vector in zip(layer_wise_weight, self.task_vectors):
            task_vector_items = list(task_vector.items())
            for w, (name, param) in zip(weight, task_vector_items):
                state_dict[name] = state_dict[name] + param * w
        self._merged_state_dict = state_dict

        return state_dict

    def forward(self, *args, **kwargs):
        if self._merged_state_dict is None:
            self.merge_weights()
        return self.forward_model(args=args, kwargs=kwargs)

    # def __getattr__(self, name: str) -> Any:
    #     try:
    #         return super().__getattr__(name)
    #     except AttributeError:
    #         attr = getattr(self.model, name)
    #         if isinstance(attr, Callable):
    #             warnings.warn(
    #                 f"forwarding `{name}` to the underlying model", UserWarning
    #             )
    #         return attr

    # def __setattr__(self, name: str, value: Any) -> None:
    #     try:
    #         super().__setattr__(name, value)
    #     except AttributeError:
    #         setattr(self.model, name, value)


def merge_weights(module: nn.Module):
    """
    Merges the weights for all `LayerWiseMergedModel` instances within the given module.

    Args:
        module (nn.Module): The module to process.
    """
    if isinstance(module, LayerWiseMergedModel):
        module.merge_weights()
        return
    else:
        for submodule in module.children():
            merge_weights(submodule)


def merge_and_unload(module: nn.Module):
    """
    Merges and unloads all `LayerWiseMergedModel` instances within the given module.

    Args:
        module (nn.Module): The module to process.

    Returns:
        nn.Module: The updated module with merged weights.
    """
    if isinstance(module, LayerWiseMergedModel):
        return module.merge_and_unload()
    else:
        for name, submodule in module.named_children():
            need_merge = isinstance(submodule, LayerWiseMergedModel)
            submodule = merge_and_unload(submodule)
            if need_merge:
                setattr(module, name, submodule)
        return module


def fix_other_parts(module: nn.Module):
    """
    Sets all parameters in the module to not require gradients, except for the merge weights
    in `LayerWiseMergedModel` instances.

    Args:
        module (nn.Module): The module to process.

    Returns:
        nn.Module: The module with updated parameter requirements.
    """
    module.requires_grad_(False)
    for submodule in module.modules():
        if isinstance(submodule, LayerWiseMergedModel):
            submodule.merge_weight.requires_grad_(True)
    return module



import logging
from pathlib import Path
from typing import Sequence

import rich
import rich.syntax
import rich.tree
from hydra.core.hydra_config import HydraConfig
from lightning.fabric.utilities.rank_zero import rank_zero_only
from omegaconf import DictConfig, OmegaConf, open_dict
from rich import print
from rich.columns import Columns
from rich.console import Console
from rich.logging import RichHandler
from rich.panel import Panel
from rich.prompt import Prompt
from rich.syntax import Syntax
from rich.text import Text

from fusion_bench.utils import pylogger

log = pylogger.RankedLogger(__name__, rank_zero_only=True)

# List of available styles
AVAILABLE_STYLES = [
    "none",
    "bold",
    "dim",
    "italic",
    "underline",
    "blink",
    "blink2",
    "reverse",
    "conceal",
    "strike",
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan",
    "white",
    "bright_red",
    "bright_green",
    "bright_yellow",
    "bright_blue",
    "bright_magenta",
    "bright_cyan",
    "bright_white",
]


def display_available_styles():
    """Display all available styles in a grid."""
    console = Console()
    style_samples = [
        Panel(f"Style: {style}", expand=False, border_style=style)
        for style in AVAILABLE_STYLES
    ]
    console.print(Columns(style_samples, equal=True, expand=False))


def print_bordered(message, title=None, style="blue", code_style=None):
    """
    Print a message with a colored border.

    Args:
    message (str): The message to print.
    title (str, optional): The title of the panel. Defaults to None.
    style (str, optional): The color style for the border. Defaults to "cyan".
    code_style (str, optional): The syntax highlighting style if the message is code.
                                Set to None for plain text. Defaults to "python".
    """
    if code_style:
        content = Syntax(message, code_style, theme="monokai", word_wrap=True)
    else:
        content = Text(message)

    panel = Panel(content, title=title, border_style=style)
    print(panel)


@rank_zero_only
def print_config_tree(
    cfg: DictConfig,
    print_order: Sequence[str] = (
        "data",
        "model",
        "callbacks",
        "logger",
        "trainer",
        "paths",
        "extras",
    ),
    resolve: bool = False,
    save_to_file: bool = False,
) -> None:
    """Prints the contents of a DictConfig as a tree structure using the Rich library.

    :param cfg: A DictConfig composed by Hydra.
    :param print_order: Determines in what order config components are printed. Default is ``("data", "model",
    "callbacks", "logger", "trainer", "paths", "extras")``.
    :param resolve: Whether to resolve reference fields of DictConfig. Default is ``False``.
    :param save_to_file: Whether to export config to the hydra output folder. Default is ``False``.
    """
    style = "tree"
    tree = rich.tree.Tree("CONFIG", style=style, guide_style=style)

    queue = []

    # add fields from `print_order` to queue
    for field in print_order:
        (
            queue.append(field)
            if field in cfg
            else log.warning(
                f"Field '{field}' not found in config. Skipping '{field}' config printing..."
            )
        )

    # add all the other fields to queue (not specified in `print_order`)
    for field in cfg:
        if field not in queue:
            queue.append(field)

    # generate config tree from queue
    for field in queue:
        branch = tree.add(field, style=style, guide_style=style)

        config_group = cfg[field]
        if isinstance(config_group, DictConfig):
            branch_content = OmegaConf.to_yaml(config_group, resolve=resolve)
        else:
            branch_content = str(config_group)

        branch.add(rich.syntax.Syntax(branch_content, "yaml"))

    # print config tree
    rich.print(tree)

    # save config tree to file
    if save_to_file:
        with open(Path(cfg.paths.output_dir, "config_tree.log"), "w") as file:
            rich.print(tree, file=file)


@rank_zero_only
def enforce_tags(cfg: DictConfig, save_to_file: bool = False) -> None:
    """Prompts user to input tags from command line if no tags are provided in config.

    :param cfg: A DictConfig composed by Hydra.
    :param save_to_file: Whether to export tags to the hydra output folder. Default is ``False``.
    """
    if not cfg.get("tags"):
        if "id" in HydraConfig().cfg.hydra.job:
            raise ValueError("Specify tags before launching a multirun!")

        log.warning("No tags provided in config. Prompting user to input tags...")
        tags = Prompt.ask("Enter a list of comma separated tags", default="dev")
        tags = [t.strip() for t in tags.split(",") if t != ""]

        with open_dict(cfg):
            cfg.tags = tags

        log.info(f"Tags: {cfg.tags}")

    if save_to_file:
        with open(Path(cfg.paths.output_dir, "tags.log"), "w") as file:
            rich.print(cfg.tags, file=file)


if __name__ == "__main__":
    print_bordered("Hello, World!", title="Greeting", style="green", code_style=None)
    print_bordered(
        "def hello_world():\n    print('Hello, World!')",
        title="Python Function",
        style="magenta",
        code_style="python",
    )
    print_bordered(
        "SELECT * FROM users WHERE age > 18;",
        title="SQL Query",
        style="yellow",
        code_style="sql",
    )

    print("\nAvailable Styles:")
    display_available_styles()


def setup_colorlogging(force=False, **config_kwargs):
    """
    Sets up color logging for the application.
    """
    FORMAT = "%(message)s"

    logging.basicConfig(
        level=logging.INFO,
        format=FORMAT,
        datefmt="[%X]",
        handlers=[RichHandler()],
        force=force,
        **config_kwargs,
    )



"""
This module is modified from https://github.com/locuslab/wanda.

It contains utility functions and classes for pruning neural network models using the Wanda method.
The WANDA method is a weight pruning technique that aims to reduce the number of parameters in a neural network
while maintaining its performance.
"""



from pathlib import Path
from typing import Literal

TQDM_BAR_FORMAT = "{l_bar}{bar:10}{r_bar}{bar:-10b}"
MODELS = ["ViT-B-32", "ViT-B-16", "ViT-L-14"]
OPENCLIP_CACHEDIR = Path(Path.home(), "openclip-cachedir", "open_clip").as_posix()
CACHEDIR = None

ALL_DATASETS = [
    "Cars",
    "DTD",
    "EuroSAT",
    "GTSRB",
    "MNIST",
    "RESISC45",
    "SVHN",
    "SUN397",
    "STL10",
    "OxfordIIITPet",
    "Flowers102",
    "CIFAR100",
    "PCAM",
    "FER2013",
    "CIFAR10",
    "Food101",
    "FashionMNIST",
    "RenderedSST2",
    "EMNIST",
    "KMNIST",
]

DATASETS_8 = ALL_DATASETS[:8]
DATASETS_14 = ALL_DATASETS[:14]
DATASETS_20 = ALL_DATASETS[:20]


def cleanup_dataset_name(dataset_name: str):
    return dataset_name.replace("Val", "") + "Val"


def get_zeroshot_path(root, dataset, model):
    return Path(
        root, model, cleanup_dataset_name(dataset), f"nonlinear_zeroshot.pt"
    ).as_posix()


def get_finetuned_path(root, dataset, model):
    return Path(
        root, model, cleanup_dataset_name(dataset), f"nonlinear_finetuned.pt"
    ).as_posix()


def get_single_task_accuracies_path(model):
    return Path(
        "results/single_task", model, f"nonlinear_ft_accuracies.json"
    ).as_posix()



import os
from copy import deepcopy
from typing import TYPE_CHECKING, Optional

from datasets import Dataset, load_dataset, load_from_disk
from lightning.fabric.utilities import rank_zero_only
from tqdm.auto import tqdm

from fusion_bench.utils import timeit_context

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer


def load_tokenized_stanford_shp_for_rlhf(
    tokenizer: "PreTrainedTokenizer",
    path: str = "stanfordnlp/SHP",
    split: str = "train",
    num_proc: int = 8,
    cache_path: Optional[str] = None,
):
    if cache_path is not None and os.path.isdir(cache_path):
        dataset = load_from_disk(cache_path)
        return dataset

    dataset = load_dataset(path, split=split)

    def tokenize(sample):
        """
        - history: the post title concatented to the post body (string)
        - human_ref_A: text of comment A (string)
        - human_ref_B: text of comment B (string)
        - labels: the preference label -- it is 1 if A is preferred to B; 0 if B is preferred to A. This was randomized such that the label distribution is roughly 50/50. (integer)
        """
        # Create a conversation with the post title and body, followed by comments
        conversation = [{"role": "user", "content": sample["history"]}]
        if sample["labels"] == 0:
            sample["chosen"] = deepcopy(conversation).append(
                {"role": "assistant", "content": sample["human_ref_B"]}
            )
            sample["rejected"] = deepcopy(conversation).append(
                {"role": "assistant", "content": sample["human_ref_A"]}
            )
        else:
            sample["chosen"] = deepcopy(conversation).append(
                {"role": "assistant", "content": sample["human_ref_A"]}
            )
            sample["rejected"] = deepcopy(conversation).append(
                {"role": "assistant", "content": sample["human_ref_B"]}
            )

        # apply chat template
        sample["chosen_chat"] = tokenizer.apply_chat_template(
            sample["chosen"], tokenize=False, add_generation_prompt=False
        )
        sample["rejected_chat"] = tokenizer.apply_chat_template(
            sample["rejected"], tokenize=False, add_generation_prompt=False
        )

        # tokenize the conversation
        tokenized_pos = tokenizer(sample["chosen_chat"], truncation=True)
        tokenized_neg = tokenizer(sample["rejected_chat"], truncation=True)

        # Ensure that the chosen response does not contain an EOS token
        sample["chosen_input_ids"] = tokenized_pos["input_ids"]
        sample["chosen_attention_mask"] = tokenized_pos["attention_mask"]
        assert (
            tokenizer.eos_token_id not in tokenized_pos["input_ids"][:-1]
        ), f"Prompt contains EOS token: {sample['positive']}"
        if sample["chosen_input_ids"][-1] != tokenizer.eos_token_id:
            sample["chosen_input_ids"].append(tokenizer.eos_token_id)
            sample["chosen_attention_mask"].append(1)

        sample["rejected_input_ids"] = tokenized_neg["input_ids"]
        sample["rejected_attention_mask"] = tokenized_neg["attention_mask"]
        # Ensure that the rejected response does not contain an EOS token
        assert (
            tokenizer.eos_token_id not in tokenized_neg["input_ids"][:-1]
        ), f"Prompt contains EOS token: {sample['rejected']}"
        if sample["rejected_input_ids"][-1] != tokenizer.eos_token_id:
            sample["rejected_input_ids"].append(tokenizer.eos_token_id)
            sample["rejected_attention_mask"].append(1)

        return sample

    dataset = dataset.map(tokenize, num_proc=num_proc)

    if cache_path is not None and rank_zero_only.rank == 0:
        dataset.save_to_disk(cache_path)
    return dataset



classnames = [
    "annual crop land",
    "forest",
    "brushland or shrubland",
    "highway or road",
    "industrial buildings or commercial buildings",
    "pasture land",
    "permanent crop land",
    "residential buildings or homes or apartments",
    "river",
    "lake or sea",
]

templates = [
    lambda c: f"a centered satellite photo of {c}.",
    lambda c: f"a centered satellite photo of a {c}.",
    lambda c: f"a centered satellite photo of the {c}.",
]



"Constants for CLIP Vision Model Merging"

TASK_NAMES_TA8 = [
    "sun397",
    "stanford-cars",
    "resisc45",
    "eurosat",
    "svhn",
    "gtsrb",
    "mnist",
    "dtd",
]
"The 8 tasks used in the Task Arithmetic paper."
TASK_NAMES_TALL8 = TASK_NAMES_TA8
"The 8 tasks used in the Tall Mask paper"
TASK_NAMES_TALL10 = TASK_NAMES_TA8 + ["oxford_flowers102", "pcam"]
TASK_NAMES_TALL12 = TASK_NAMES_TALL10 + [
    "fer2013",
    "oxford-iiit-pet",
]
TASK_NAMES_TALL14 = TASK_NAMES_TALL12 + [
    "stl10",
    "cifar100",
]
"The 14 tasks used in the TALL mask paper"
TASK_NAMES_TALL16 = TASK_NAMES_TALL14 + ["cifar10", "food101"]
TASK_NAMES_TALL18 = TASK_NAMES_TALL16 + ["fashion_mnist", "emnist_letters"]
TASK_NAMES_TALL20 = TASK_NAMES_TALL18 + ["kmnist", "rendered-sst2"]
"The 20 tasks used in the TALL mask paper"
TASK_NAMES_TA8_CAP = [
    "SUN397",
    "Cars",
    "RESISC45",
    "EuroSAT",
    "SVHN",
    "GTSRB",
    "MNIST",
    "DTD",
]
TASK_NAMES_TALL8_CAP = TASK_NAMES_TA8_CAP
TASK_NAMES_TALL10_CAP = TASK_NAMES_TALL8_CAP + ["Flowers102", "PCAM"]
TASK_NAMES_TALL12_CAP = TASK_NAMES_TALL10_CAP + ["FER2013", "OxfordIIITPet"]
TASK_NAMES_TALL14_CAP = TASK_NAMES_TALL12_CAP + ["STL10", "CIFAR100"]
TASK_NAMES_TALL16_CAP = TASK_NAMES_TALL14_CAP + ["CIFAR10", "Food101"]
TASK_NAMES_TALL18_CAP = TASK_NAMES_TALL16_CAP + ["FashionMNIST", "EMNIST"]
TASK_NAMES_TALL20_CAP = TASK_NAMES_TALL18_CAP + ["KMNIST", "RenderedSST2"]



"""
Base Program Classes for FusionBench.

This module defines the foundational abstract base classes for FusionBench programs.
These programs serve as the main execution units that orchestrate model fusion
workflows, from loading configurations to executing fusion algorithms and
evaluating results.

The base classes provide a consistent interface for all FusionBench programs
while allowing for flexible implementations of different fusion workflows.
"""

from abc import abstractmethod

from fusion_bench.mixins import BaseYAMLSerializable


class BaseHydraProgram(BaseYAMLSerializable):
    """
    Abstract base class for all FusionBench programs that use Hydra configuration.

    This class serves as the foundation for all FusionBench execution programs,
    providing a standardized interface for configuration-driven model fusion
    workflows. It combines the serialization capabilities of BaseYAMLSerializable
    with the requirement for a main execution method.

    The class is designed to work seamlessly with Hydra's configuration management
    system, allowing programs to be instantiated and configured through YAML files.
    This enables flexible, reproducible experiments with different fusion algorithms,
    model pools, and evaluation tasks.

    Key Features:

    - Configuration-driven execution through Hydra integration
    - YAML serialization support for experiment reproducibility
    - Abstract interface ensuring consistent program structure
    - Integration with FusionBench's modular architecture

    Typical Usage:
        Subclasses should implement the `run()` method to define their specific
        fusion workflow. The program can then be executed through the FusionBench
        CLI or instantiated directly from configuration files.

    Example:
        ```python
        class MyFusionProgram(BaseHydraProgram):
            def __init__(self, method_config, modelpool_config, taskpool_config):
                self.method_config = method_config
                self.modelpool_config = modelpool_config
                self.taskpool_config = taskpool_config

            def run(self):
                # Load components
                algorithm = load_algorithm(self.method_config)
                modelpool = load_modelpool(self.modelpool_config)
                taskpool = load_taskpool(self.taskpool_config)

                # Execute fusion
                merged_model = algorithm.run(modelpool)

                # Evaluate results
                report = taskpool.evaluate(merged_model)
                return report
        ```

    Note:
        This is an abstract base class and cannot be instantiated directly.
        Subclasses must implement the `run()` method to provide concrete
        functionality.

    See Also:

    - [FabricModelFusionProgram][fusion_bench.programs.FabricModelFusionProgram]: Lightning Fabric-based implementation
    - [BaseYAMLSerializable][fusion_bench.mixins.BaseYAMLSerializable]: Parent class providing serialization
    - FusionBench CLI documentation for program execution details
    """

    @abstractmethod
    def run(self):
        """
        Execute the main program workflow.

        This abstract method defines the primary entry point for program execution.
        Subclasses must implement this method to define their specific fusion
        workflow, including model loading, fusion algorithm execution, and
        result evaluation.
        """
        pass



import logging
from collections import defaultdict
from typing import Dict, List, cast  # noqa: F401

import torch
import torch.utils.data
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.nn.modules import Module
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm

from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import CLIPClassificationMixin

from .regmean_plusplus import RegMeanAlgorithmPlusPlus

log = logging.getLogger(__name__)


class RegMeanAlgorithmForCLIPPlusPlus(
    RegMeanAlgorithmPlusPlus,
    CLIPClassificationMixin,
):
    _config_mapping = {
        "_dataloader_kwargs": "dataloader_kwargs",
    }

    def __init__(self, *, dataloader_kwargs: DictConfig, **kwargs):
        super().__init__(**kwargs)
        self.dataloader_kwargs = dataloader_kwargs

    def on_regmean_start(self):
        self.setup_zero_shot_classification_head()

    def compute_logits(self, module, batch, task: str) -> Tensor:
        images, _ = batch
        text_embeds = self.zeroshot_weights[task]

        image_embeds = module(images)[1]
        image_embeds = self.visual_projection(image_embeds)

        # normalize embeddings
        image_embeds = image_embeds / image_embeds.norm(p=2, dim=-1, keepdim=True)

        # cosine similarity
        logits_per_text = (
            torch.matmul(text_embeds, image_embeds.t()) * self.logit_scale_exp
        )
        logits_per_image = logits_per_text.t()

        return logits_per_image

    def get_regmean_weights(
        self,
        model_name: str,
        layer: Module,
        batches_input: List[Tensor],
        linear_modules_to_merge: Dict[str, Module],
    ):
        layer = self.fabric.setup(layer)

        def compute_regmean_weights(module_name: str):
            """
            compute the regmean weights, a hook function to deal with each module's input
            :param module_name: str, module name
            :return:
            """

            def hook(module: nn.Module, input: tuple, output: torch.Tensor):
                # Tensor, shape (batch_size, sequence_length, hidden_dim)
                x = cast(Tensor, input[0]).detach()
                batch_num_actual_examples = x.shape[0]
                # Tensor, shape (batch_size * sequence_length, hidden_dim)
                x = x.reshape(-1, x.shape[-1])
                # Tensor, shape (hidden_dim, hidden_dim)
                xtx = torch.matmul(x.transpose(0, 1), x)
                # store the averaged weights in regmean_weights
                if module_name not in regmean_weights.keys():
                    regmean_weights[module_name] = xtx / x.shape[0]
                    num_computed_examples[module_name] = x.shape[0]
                    num_actual_examples[module_name] = batch_num_actual_examples
                else:
                    regmean_weights[module_name] = (
                        regmean_weights[module_name]
                        * num_computed_examples[module_name]
                        + xtx
                    ) / (num_computed_examples[module_name] + x.shape[0])
                    num_computed_examples[module_name] += x.shape[0]
                    num_actual_examples[module_name] += batch_num_actual_examples

            return hook

        handles = []
        # dictionary, regmean matrices for each linear module inputs
        regmean_weights = {}
        # dictionary, number of examples (multiplied the sequence length) used for computing regmean matrices
        num_computed_examples = {}
        # dictionary, number of actual examples used for computing regmean matrices
        num_actual_examples = {}

        for module_name, linear_module_to_merge in linear_modules_to_merge.items():
            # register a hook in the forward process
            handle = linear_module_to_merge.register_forward_hook(
                compute_regmean_weights(module_name=module_name)
            )
            handles.append(handle)
        _ = self.layer_batches_forward(layer, batches_input)

        # remove the added hook
        for handle in handles:
            handle.remove()

        for module_name in regmean_weights.keys():
            regmean_weights[module_name] = regmean_weights[module_name].detach().cpu()

        return regmean_weights

    def merge_embedding_layer(self, models_to_merge_dict: Dict[str, nn.Module]):
        models_to_merge_param_dict = defaultdict(list)

        # get the parameters of the embedding layer from each model
        for model_to_merge in models_to_merge_dict.values():
            model_to_merge_state_dict = model_to_merge.state_dict()

            param_dict = {}
            for name, param in model_to_merge_state_dict.items():
                if name.startswith("vision_model.embeddings") or name.startswith(
                    "vision_model.pre_layrnorm"
                ):
                    param_dict[name] = param

            for param_name in param_dict.keys():
                models_to_merge_param_dict[param_name].append(param_dict[param_name])

        # merge the parameters of the embedding layer
        merged_params_dict = {}
        for param_name, param_list in models_to_merge_param_dict.items():
            merged_params_dict[param_name] = torch.stack(param_list).mean(dim=0)

        return merged_params_dict

    def get_input_for_first_layer(self, model: nn.Module, train_dataset):
        # setup dataloader
        train_dataset = CLIPDataset(train_dataset, self.clip_processor)
        train_dataloader = DataLoader(
            train_dataset, shuffle=True, **self.dataloader_kwargs
        )
        train_dataloader = self.fabric.setup_dataloaders(train_dataloader)
        model = self.fabric.setup(model)

        def compute_input(model, batch):
            images, _ = batch

            images = images.to(model.device)
            image_embeds = model.vision_model.embeddings(images)
            image_embeds = model.vision_model.pre_layrnorm(image_embeds)
            image_embeds = image_embeds.detach().cpu()

            return image_embeds

        num_computed_examples = 0
        num_regmean_examples = self.num_regmean_examples

        batches_input = []
        for batch in train_dataloader:
            if num_computed_examples >= num_regmean_examples:
                break
            batches_input.append(compute_input(model, batch))
            num_computed_examples += batch[0].size(0)

        return batches_input

    def get_layers(self, model: nn.Module):
        return model.vision_model.encoder.layers

    def update_merged_params_dict(
        self, merged_params_dict, new_merged_params, layer_idx
    ):
        for key, value in new_merged_params.items():
            key = f"vision_model.encoder.layers.{layer_idx}.{key}"
            merged_params_dict[key] = value

        return merged_params_dict

    def layer_batches_forward(
        self, layer: nn.Module, batches_input: List[Tensor]
    ) -> Tensor:
        batches_output = []
        for batch in batches_input:
            device = next(layer.parameters()).device
            batch = batch.to(device)
            logits = (
                layer(batch, attention_mask=None, causal_attention_mask=None)[0]
                .detach()
                .cpu()
            )
            batches_output.append(logits)
        return batches_output



import logging
from typing import List, Optional, Union, cast  # noqa: F401

import torch
from omegaconf import open_dict
from transformers import (
    LlamaForCausalLM,
    LlamaModel,
    MistralForCausalLM,
    MistralModel,
    MixtralForCausalLM,
    MixtralModel,
)
from transformers.models.llama.modeling_llama import LlamaDecoderLayer
from transformers.models.mistral.modeling_mistral import MistralDecoderLayer
from transformers.models.mixtral.modeling_mixtral import MixtralDecoderLayer

from fusion_bench.modelpool import BaseModelPool

from .mixtral_upcycling import (
    MixtralForCausalLMUpscalingAlgorithm,
    MixtralUpscalingAlgorithm,
    _convert_mlp,
)

log = logging.getLogger(__name__)


def _substitute_experts(
    expert_idx: int,
    expert_model: Union[LlamaModel, MistralModel],
    mixtral_model: MixtralModel,
):
    """
    Substitute the experts of the `MixtralModel` with the models from the modelpool.

    Args:
        expert_idx (int): The index of the expert to substitute.
        expert_model (Union[LlamaModel, MistralModel]): The expert model to substitute.
        mixtral_model (MixtralModel): The MixtralModel to substitute the experts in.
    """
    for input_layer, output_layer in zip(expert_model.layers, mixtral_model.layers):
        output_layer = cast(MixtralDecoderLayer, output_layer)
        input_layer = cast(Union[LlamaDecoderLayer, MistralDecoderLayer], input_layer)
        _convert_mlp(input_layer.mlp, output_layer.block_sparse_moe.experts[expert_idx])


class MixtralMoEMergingAlgorithm(MixtralUpscalingAlgorithm):
    """
    This class is responsible for merging models into a MixtralModel.
    """

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool) -> MixtralModel:
        """
        Runs the merging process.

        Args:
            modelpool (ModelPool): The pool of models to be merged. Each model in the pool will be treated as an expert, and should be a `MistralModel` or `LlamaModel`.

        Returns:
            MixtralModel: The merged model.
        """
        with open_dict(self.config):
            self.config.num_experts = len(modelpool)

        # firstly, we upscale the models to MixtralModel
        mixtral_model = super()._run(modelpool)

        # then we substitute the experts of the MixtralModel with the models from the modelpool
        for model_idx, model_name in enumerate(modelpool.model_names):
            expert_model: MistralModel | LlamaModel = modelpool.load_model(model_name)
            _substitute_experts(model_idx, expert_model, mixtral_model)

        if self.config.get("save_checkpoint", None) is not None:
            mixtral_model.save_pretrained(self.config.save_checkpoint)
        return mixtral_model


class MixtralForCausalLMMergingAlgorithm(MixtralForCausalLMUpscalingAlgorithm):
    """
    This class is responsible for merging models into a `MixtralForCausalLM`.
    """

    @torch.no_grad()
    def run(self, modelpool: BaseModelPool) -> MixtralForCausalLM:
        """
        Runs the merging process. It first upscales the models to MixtralForCausalLM,
        then substitutes the experts of the MixtralForCausalLM with the models from the modelpool.

        Args:
            modelpool (ModelPool): The pool of models to be merged. Each model in the pool will be treated as an expert, and should be a `MistralForCausalLM` or `LlamaForCausalLM`.

        Returns:
            MixtralForCausalLM: The merged model.
        """
        with open_dict(self.config):
            self.config.num_experts = len(modelpool)

        # firstly, we upscale the models to MixtralForCausalLM
        mixtral_model = super()._run(modelpool)

        # then we substitute the experts of the MixtralForCausalLM with the models from the modelpool
        for model_idx, model_name in enumerate(modelpool.model_names):
            expert_model: MistralForCausalLM | LlamaForCausalLM = modelpool.load_model(
                model_name
            )
            _substitute_experts(model_idx, expert_model.model, mixtral_model.model)

        if self.config.get("save_checkpoint", None) is not None:
            mixtral_model.save_pretrained(self.config.save_checkpoint)
        return mixtral_model



"""
This module provides utilities for visualizing learning rate schedulers.

Functions:
    simulate_scheduler(lr_scheduler, steps): Simulates the learning rate scheduler for a given number of steps.
    plot_lr_schedulers(lr_schedulers, steps, titles): Plots the learning rates of one or more schedulers over a number of steps.
"""

from typing import TYPE_CHECKING, List, Union

import matplotlib.pyplot as plt
import torch

if TYPE_CHECKING:
    from torch.optim.lr_scheduler import LRScheduler

__all__ = ["simulate_scheduler", "plot_lr_schedulers"]


def simulate_scheduler(lr_scheduler, steps: int):
    """
    Simulates the learning rate scheduler for a given number of steps.

    Args:
        lr_scheduler (torch.optim.lr_scheduler.LRScheduler): The learning rate scheduler object.
        steps (int): The number of steps to simulate.

    Returns:
        List[float]: A list of learning rates for each step.
    """
    lrs = []
    for _ in range(steps):
        lr = lr_scheduler.step()
        lrs.append(lr)
    return lrs


def plot_lr_schedulers(
    lr_schedulers: Union["LRScheduler", List["LRScheduler"]],
    steps: int,
    titles: Union[str, List[str]],
    show_plot: bool = True,
):
    """
    Plots the learning rates of one or more schedulers over a number of steps.

    Args:
        lr_schedulers (Union[LRScheduler, List[LRScheduler]]): One or more learning rate scheduler objects.
        steps (int): The number of steps to simulate.
        titles (Union[str, List[str]]): Titles for the plots.

    Returns:
        fig, axes: The matplotlib figure and axes objects.
    """
    # Handle single scheduler
    if isinstance(lr_schedulers, torch.optim.lr_scheduler.LRScheduler):
        lr_schedulers = [lr_schedulers]
    if isinstance(titles, str):
        titles = [titles]

    fig, axs = plt.subplots(len(lr_schedulers), 1, figsize=(5, 3 * len(lr_schedulers)))
    if len(lr_schedulers) == 1:
        axs = [axs]

    for i, (scheduler, title) in enumerate(zip(lr_schedulers, titles)):
        lrs = simulate_scheduler(scheduler, steps)
        axs[i].plot(lrs, label=title)
        axs[i].set_title(title)
        axs[i].set_xlabel("Steps")
        axs[i].set_ylabel("Learning Rate")
        axs[i].legend()
        axs[i].grid(True)

    plt.tight_layout()
    if show_plot:
        plt.show()
    return fig, axs


# Example usage
if __name__ == "__main__":
    from fusion_bench.optim.lr_scheduler.linear_warmup import (
        CosineDecayWithWarmup,
        LinearWarmupScheduler,
        PolySchedulerWithWarmup,
    )

    # Dummy optimizer
    optimizer = torch.optim.SGD(
        [torch.nn.Parameter(torch.randn(2, 2, requires_grad=True))], lr=0.1
    )

    # Define the schedulers
    linear_scheduler = LinearWarmupScheduler(
        optimizer, t_max=100, max_lr=0.1, min_lr=0.01, init_lr=0.0, warmup_steps=10
    )
    cosine_scheduler = CosineDecayWithWarmup(
        optimizer, t_max=100, max_lr=0.1, min_lr=0.01, init_lr=0.0, warmup_steps=10
    )
    poly_scheduler = PolySchedulerWithWarmup(
        optimizer,
        t_max=100,
        max_lr=0.1,
        min_lr=0.01,
        init_lr=0.0,
        warmup_steps=40,
        poly_order=2.0,
    )

    # Plot the learning rates
    plot_lr_schedulers(
        [linear_scheduler, cosine_scheduler, poly_scheduler],
        steps=100,
        titles=[
            "Linear Warmup",
            "Cosine Decay with Warmup",
            "Poly Scheduler with Warmup",
        ],
    )



import logging
import os
from typing import Any, Dict, List, Optional

from datasets import load_dataset, load_from_disk
from transformers import PreTrainedTokenizer
from trl import SFTConfig, SFTTrainer

import fusion_bench

log = logging.getLogger(__name__)



import torch
from datasets import load_dataset
from transformers import default_data_collator


def _preprocess(tokenizer, examples, max_length=128):
    return tokenizer(
        examples["text"], padding="max_length", truncation=True, max_length=max_length
    )


def get_dataset(dataset_name: str, subset: str, split: str, size: int = None):
    if size is None:
        dataset = load_dataset(dataset_name, subset)[split]
    else:
        dataset = load_dataset(dataset_name, subset, streaming=True)[split]
        dataset = dataset.take(size)

    return dataset


def get_dataloader(dataset, tokenizer, batch_size, num_workers=4, max_length=128):
    dataset = dataset.map(
        lambda examples: _preprocess(tokenizer, examples, max_length),
        batched=True,
        batch_size=batch_size,
        remove_columns=["text", "timestamp", "url"],
    )
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        num_workers=0,
        collate_fn=default_data_collator,
    )
    return dataloader



import logging
from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union, cast

import lightning.fabric
import numpy as np
import pandas as pd
import torch
import torch.nn.functional as F
from omegaconf import DictConfig, OmegaConf
from open_clip.model import ResidualAttentionBlock
from torch import Tensor, nn
from tqdm.auto import tqdm

from fusion_bench import BaseAlgorithm
from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.method.task_arithmetic import task_arithmetic_merge
from fusion_bench.mixins import OpenCLIPClassificationMixin, SimpleProfilerMixin
from fusion_bench.modelpool import OpenCLIPVisionModelPool
from fusion_bench.models.open_clip import ClassificationHead, ImageEncoder
from fusion_bench.utils import print_parameters, timeit_context
from fusion_bench.utils.data import InfiniteDataLoader

from .module import ParetoWeightEnsemblingModule
from .phn.solvers import EPOSolver
from .utils import generate_simplex_grid

log = logging.getLogger(__name__)


class PWEMoEAlgorithmForOpenCLIP(
    BaseAlgorithm,
    SimpleProfilerMixin,
    OpenCLIPClassificationMixin,
):
    modelpool: OpenCLIPVisionModelPool

    #! === Training & Validation Data ===
    # setup the datasets and loaders by calling `load_datasets`
    train_datasets: Dict[str, CLIPDataset]
    train_loaders: Dict[str, torch.utils.data.DataLoader]
    train_loader_iters: Dict[str, Iterator[Tuple[torch.Tensor, torch.Tensor]]]

    test_datasets: Dict[str, CLIPDataset]
    test_loaders: Dict[str, torch.utils.data.DataLoader]

    def __init__(
        self,
        *,
        #! === Model Architecture Arguments ===
        partial: bool,
        init_lambda: float,
        router_hidden_layers: int,
        checkpoint_path: str,
        #! === Training Arguments ===
        run_train: bool,
        num_steps: int,
        save_interval: int,
        lr: float,
        alpha: float,
        dataloader_kwargs: DictConfig,
        #! === Evaluation Arguments ===
        run_eval: bool,
        num_evaluation_samples: Union[str, int],
        quick_evaluation: bool,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.partial = partial
        self.init_lambda = init_lambda
        self.router_hidden_layers = router_hidden_layers
        self.lr = lr
        self.num_steps = num_steps
        self.save_interval = save_interval
        self.alpha = alpha
        self.checkpoint_path = checkpoint_path
        self._dataloader_kwargs = dataloader_kwargs
        self.run_train = run_train
        self.run_eval = run_eval
        self.num_evaluation_samples = num_evaluation_samples
        self.quick_evaluation = quick_evaluation

    def run(self, modelpool: OpenCLIPVisionModelPool):
        self.modelpool = modelpool

        # setup the MoE model
        model = self.load_model()
        if self.checkpoint_path is not None:
            self.fabric.load(self.checkpoint_path, {"model": model})

        # setup dataloaders
        self.load_datasets()

        if self.run_train:
            model = self.train()
        if self.run_eval:
            self.evaluate(model)
        return model

    @torch.no_grad()
    def load_model(self):
        modelpool = self.modelpool

        # load models and classification heads
        pretrained_model: ImageEncoder = self.modelpool.load_pretrained_model()
        log.info("pretrained model statistics:")
        print_parameters(pretrained_model, print_fn=log.info)

        finetuned_models: Dict[str, ImageEncoder] = {}
        for model_name in self.modelpool.model_names:
            finetuned_models[model_name] = modelpool.load_model(model_name)

        classification_heads: Dict[str, ClassificationHead] = {}
        for model_name in self.modelpool.model_names:
            classification_heads[model_name] = modelpool.load_classification_head(
                model_name
            )
        self.classification_heads = classification_heads

        self.train_processor = modelpool.train_processor
        self.test_processor = modelpool.test_processor

        with timeit_context("Building the MoE model"):
            model = deepcopy(pretrained_model)

            if self.partial:
                log.info("Weight ensembling only the MLPs")
                # weight ensembling only the MLPs, merge the remaining layers using task arithmetic
                model = task_arithmetic_merge(
                    pretrained_model=model,
                    finetuned_models=list(finetuned_models.values()),
                    scaling_factor=self.init_lambda,
                    inplace=True,
                )

                # fix all parameters
                model.requires_grad_(False)

                for layer_idx in tqdm(
                    range(model.model.visual.transformer.layers), desc="Upscaling MLPs"
                ):
                    resblock: ResidualAttentionBlock = (
                        model.model.visual.transformer.resblocks[layer_idx]
                    )
                    resblock.mlp = ParetoWeightEnsemblingModule(
                        base_model=cast(
                            ResidualAttentionBlock,
                            pretrained_model.model.visual.transformer.resblocks[
                                layer_idx
                            ],
                        ).mlp,
                        expert_models=[
                            cast(
                                ResidualAttentionBlock,
                                m.model.visual.transformer.resblocks[layer_idx],
                            ).mlp
                            for m in finetuned_models.values()
                        ],
                        init_lambda=self.init_lambda,
                        fix_base_model_and_experts=True,
                        router_hidden_layers=self.router_hidden_layers,
                    )
            else:
                log.info("Weight ensembling all the layers")
                # weight ensembling all the layers, merge the remaining layers using task arithmetic
                model = task_arithmetic_merge(
                    pretrained_model=model,
                    finetuned_models=list(finetuned_models.values()),
                    scaling_factor=self.init_lambda,
                    inplace=True,
                )
                # fix all parameters
                model.requires_grad_(False)

                for name in [
                    "conv1",
                    "ln_pre",
                    "ln_post",
                    # "class_embedding",
                    # "positional_embedding",
                ]:
                    setattr(
                        model.model.visual,
                        name,
                        ParetoWeightEnsemblingModule(
                            base_model=getattr(pretrained_model.model.visual, name),
                            expert_models=[
                                getattr(m.model.visual, name)
                                for m in finetuned_models.values()
                            ],
                            init_lambda=self.init_lambda,
                            fix_base_model_and_experts=True,
                            router_hidden_layers=self.router_hidden_layers,
                        ),
                    )
                for layer_idx in tqdm(
                    range(model.model.visual.transformer.layers),
                    desc="Upscaling the transformer layers",
                ):
                    for name in ["ln_1", "attn", "ln_attn", "ln_2", "mlp"]:
                        setattr(
                            model.model.visual.transformer.resblocks[layer_idx],
                            name,
                            ParetoWeightEnsemblingModule(
                                base_model=getattr(
                                    cast(
                                        ResidualAttentionBlock,
                                        pretrained_model.model.visual.transformer.resblocks[
                                            layer_idx
                                        ],
                                    ),
                                    name,
                                ),
                                expert_models=[
                                    getattr(
                                        cast(
                                            ResidualAttentionBlock,
                                            m.model.visual.transformer.resblocks[
                                                layer_idx
                                            ],
                                        ),
                                        name,
                                    )
                                    for m in finetuned_models.values()
                                ],
                                init_lambda=self.init_lambda,
                                fix_base_model_and_experts=True,
                                router_hidden_layers=self.router_hidden_layers,
                            ),
                        )
                for name in ["token_embedding", "ln_final"]:
                    setattr(
                        model.model,
                        name,
                        ParetoWeightEnsemblingModule(
                            base_model=getattr(pretrained_model.model, name),
                            expert_models=[
                                getattr(m.model, name)
                                for m in finetuned_models.values()
                            ],
                            init_lambda=self.init_lambda,
                            fix_base_model_and_experts=True,
                            router_hidden_layers=self.router_hidden_layers,
                        ),
                    )

            self.model = model
            print_parameters(model, print_fn=log.info)
            return model

    def load_datasets(self):
        modelpool = self.modelpool

        # setup the train datasets and loaders
        train_datasets = {}
        train_loaders = {}
        train_loader_iters = {}
        for dataset_name in modelpool.train_dataset_names:
            train_datasets[dataset_name] = modelpool.load_train_dataset(dataset_name)
            train_datasets[dataset_name] = CLIPDataset(
                train_datasets[dataset_name], self.train_processor
            )
            # sanity check
            assert isinstance(train_datasets[dataset_name][0], tuple)

            # setup the train loaders
            train_loaders[dataset_name] = torch.utils.data.DataLoader(
                train_datasets[dataset_name],
                shuffle=True,
                drop_last=True,
                **self._dataloader_kwargs,
            )
            train_loaders[dataset_name] = self.fabric.setup_dataloaders(
                train_loaders[dataset_name]
            )
            train_loaders[dataset_name] = InfiniteDataLoader(
                train_loaders[dataset_name]
            )

            # setup the train loader iterators
            train_loader_iters[dataset_name] = iter(train_loaders[dataset_name])

        self.train_datasets = train_datasets
        self.train_loaders = train_loaders
        self.train_loader_iters = train_loader_iters

        # setup the test datasets and loaders
        test_datasets = {}
        test_loaders = {}
        for dataset_name in modelpool.test_dataset_names:
            test_datasets[dataset_name] = modelpool.load_test_dataset(dataset_name)
            test_datasets[dataset_name] = CLIPDataset(
                test_datasets[dataset_name], self.test_processor
            )
            test_loaders[dataset_name] = torch.utils.data.DataLoader(
                test_datasets[dataset_name],
                shuffle=False,
                **self._dataloader_kwargs,
            )
            test_loaders[dataset_name] = self.fabric.setup_dataloaders(
                test_loaders[dataset_name]
            )

        self.test_datasets = test_datasets
        self.test_loaders = test_loaders

    def compute_loss(self, model: ImageEncoder, ray: Tensor):
        losses = []
        for dataset_idx, dataset_name in enumerate(self.train_datasets):
            batch = next(self.train_loader_iters[dataset_name])
            x, y = batch

            features = model(x)
            logits = self.classification_heads[dataset_name](features)

            _loss = F.cross_entropy(logits, y)
            losses.append(_loss)

        loss = self.aggregate_loss(model, ray, losses)
        return loss

    @abstractmethod
    def aggregate_loss(self, model: nn.Module, ray: Tensor, losses: Tuple[Tensor]):
        pass

    def train(self):
        # setup the model
        num_objectives = len(self.modelpool.model_names)
        model = deepcopy(self.model)
        self.classification_heads = {
            t: h.to(self.fabric.device) for t, h in self.classification_heads.items()
        }

        # set up the optimizer and learning rate scheduler
        optimizer = torch.optim.Adam(
            filter(lambda p: p.requires_grad, model.parameters()),
            lr=self.lr,
        )
        model, optimizer = self.fabric.setup(model, optimizer)
        lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer=optimizer, T_max=self.num_steps, eta_min=self.lr * 0.1
        )

        model.train()
        device = self.fabric.device
        for step_idx in tqdm(
            range(1, 1 + self.num_steps), "training", dynamic_ncols=True
        ):
            # sample a preference ray
            ray = torch.from_numpy(
                np.random.dirichlet((self.alpha,) * num_objectives, 1)
                .astype(np.float32)
                .flatten()
            ).to(device)
            ParetoWeightEnsemblingModule.set_preferenece_vector(model, ray)

            loss = self.compute_loss(model, ray)

            optimizer.zero_grad()
            self.fabric.backward(loss)
            optimizer.step()

            lr_scheduler.step()

            self.fabric.log("loss", loss.item(), step=step_idx)

            if step_idx % self.save_interval == 0 or step_idx == self.num_steps:
                ckpt_dir = Path(self.log_dir) / "checkpoints"
                ckpt_dir.mkdir(exist_ok=True, parents=True)
                self.fabric.save(
                    ckpt_dir / f"model_step={step_idx}.ckpt",
                    {"model": model},
                )
        return model

    def evaluate(self, model):
        results = defaultdict(list)

        num_objectives = len(self.modelpool.model_names)
        device = self.fabric.device
        self.classification_heads = {
            t: h.to(self.fabric.device) for t, h in self.classification_heads.items()
        }

        if not lightning.fabric.is_wrapped(model):
            model = self.fabric.setup_module(model)
        model.eval()

        if self.num_evaluation_samples == "equal_weight":
            uniform_grid = np.array(
                [[1 / num_objectives] * num_objectives], dtype=np.float32
            )
        else:
            uniform_grid = generate_simplex_grid(
                num_objectives, self.num_evaluation_samples
            )
        for ray_idx, ray in tqdm(enumerate(uniform_grid), "evaluating samples"):
            results["ray_idx"].append(ray_idx)
            # sample a preference ray
            for i in range(len(ray)):
                results[f"ray_{i}"].append(ray[i])
            ray = torch.from_numpy(ray).to(device)
            ParetoWeightEnsemblingModule.set_preferenece_vector(model, ray)

            accs = []
            for dataset_idx, dataset_name in enumerate(
                tqdm(
                    self.modelpool.test_dataset_names,
                    "evaluating datasets",
                    leave=False,
                )
            ):
                test_loader = self.test_loaders[dataset_name]
                TOTAL_CORRECT = 0
                TOTAL_COUNT = 0
                for batch_idx, batch in enumerate(
                    pbar := tqdm(
                        test_loader,
                        f"evaluate {dataset_name}",
                        leave=False,
                    )
                ):
                    x, y = batch

                    features = model(x)
                    logits = self.classification_heads[dataset_name](features)
                    preds = logits.argmax(-1)

                    correct = (preds == y).sum().item()
                    TOTAL_CORRECT += correct
                    TOTAL_COUNT += len(y)
                    acc = TOTAL_CORRECT / TOTAL_COUNT
                    pbar.set_postfix_str(f"acc={acc:.2f}")

                    if self.quick_evaluation and batch_idx > 20:
                        break
                results[dataset_name].append(acc)
                accs.append(acc)

            # compute the average accuracy
            if "average" not in self.modelpool.test_dataset_names:
                results["average"].append(np.mean(accs))

            (df := pd.DataFrame(results)).to_csv(
                Path(self.log_dir) / "result.csv", index=False
            )
            log.info(df)


class PWEMoELinearScalarizationForOpenCLIP(PWEMoEAlgorithmForOpenCLIP):
    def aggregate_loss(self, model: nn.Module, ray: Tensor, losses: Tuple[Tensor]):
        loss = 0
        for r, l in zip(ray, losses):
            loss += r * l
        return loss


class PWEMoEExactParetoOptimalForOpenCLIP(PWEMoEAlgorithmForOpenCLIP):
    epo_solver: Optional[EPOSolver] = None

    def aggregate_loss(self, model: nn.Module, ray: Tensor, losses: Tuple[Tensor]):
        if self.epo_solver is None:
            num_objectives = len(self.modelpool.model_names)
            self.epo_solver = EPOSolver(n_tasks=num_objectives, n_params=None)
        epo_solver = self.epo_solver

        losses = torch.stack(losses)
        loss = epo_solver.get_weighted_loss(
            losses,
            ray,
            tuple(filter(lambda p: p.requires_grad, model.parameters())),
        )
        return loss



from . import register
from .configuration_smile_mistral import SmileMistralConfig
from .modeling_smile_mistral import (
    SmileMistralDecoderLayer,
    SmileMistralForCausalLM,
    SmileMistralModel,
)



"""
This module contains the code for loading the calibration data.

- C4
- Math
"""

import itertools
import logging
import os

import torch
import transformers
from datasets import load_dataset
from huggingface_hub import hf_hub_download
from transformers import PreTrainedTokenizer, default_data_collator
from transformers.testing_utils import CaptureLogger

logger = logging.getLogger(__name__)


DATASETS = {
    # C4: Please download first part of the C4 training data `c4-train.00000-of-01024.json` from [allenai/c4](https://huggingface.co/datasets/allenai/c4/blob/main/en/c4-train.00000-of-01024.json.gz).
    "c4": lambda: load_dataset(
        "json",
        data_files={
            "train": hf_hub_download(
                "allenai/c4",
                filename="en/c4-train.00000-of-01024.json.gz",
                repo_type="dataset",
            )
        },
    ),
    # MATH: You can use our pre-built calibration set in `./data/math_pretrain_style.json`. To reproduce our construction, please download the training set of [MATH](https://github.com/hendrycks/math) and use our [script](data/math_calib_construction.py).
    # NOTE: I have uploaded the math_pretrain_style.json to my huggingface repo:
    # https://huggingface.co/datasets/tanganke/math_pretrain_style/tree/main.
    "math": lambda: load_dataset(
        "json",
        data_files={
            "train": hf_hub_download(
                "tanganke/math_pretrain_style",
                filename="math_pretrain_style.json",
                repo_type="dataset",
            )
        },
    ),
}


def build_calib_loader(
    dataset: str,
    tokenizer: PreTrainedTokenizer,
    max_block_size: int,
    n_blocks_for_stat: int,
    batch_size: int,
    num_workers: int,
    seed: int = 42,
):
    # dataset can be a string or a dataset object.
    # If it is a string, it can be the name of the dataset in DATASETS or the path to the dataset (a json file).
    if isinstance(dataset, str):
        if dataset in DATASETS:
            all_set = DATASETS[dataset]()
        else:
            assert os.path.exists(dataset), f"Dataset {dataset} not found."
            all_set = load_dataset("json", data_files={"train": dataset})
    else:
        assert dataset is not None, "Dataset is not provided."
        all_set = dataset

    block_size = tokenizer.model_max_length
    if block_size > max_block_size:
        logger.info(
            "The chosen tokenizer supports a `model_max_length` that is longer than the default `max_block_size` value"
            f" of {max_block_size}. If you would like to use a longer `block_size` up to `tokenizer.model_max_length` you can"
            " override this default with `--max_block_size xxx`."
        )
        block_size = max_block_size

    if n_blocks_for_stat > 0:  # Random choose `n_blocks_for_stat` blocks
        calib_set = (
            all_set["train"]
            .shuffle(seed=seed)
            .select(range(min(n_blocks_for_stat * 16, len(all_set["train"]))))
        )
    else:  # Use the whole set
        logger.warning("n_blocks_for_stat <= 0, using the whole dataset.")
        calib_set = all_set["train"].shuffle(seed=seed)

    logger.info(f"Calibration dataset: {calib_set}")
    text_column_name = (
        "text" if "text" in calib_set.features else list(calib_set.features)[0]
    )

    tok_logger = transformers.utils.logging.get_logger(
        "transformers.tokenization_utils_base"
    )

    def tokenize_function(examples):
        with CaptureLogger(tok_logger) as cl:
            output = tokenizer(examples[text_column_name])
        # clm input could be much much longer than block_size
        if "Token indices sequence length is longer than the" in cl.out:
            tok_logger.warning(
                "^^^^^^^^^^^^^^^^ Please ignore the warning above - this long input will be chunked into smaller bits"
                " before being passed to the model."
            )
        return output

    tokenized_calib_set = calib_set.map(
        tokenize_function,
        batched=True,
        remove_columns=list(calib_set.features),
    )

    def group_texts(examples):
        # Concatenate all texts.
        concatenated_examples = {
            k: list(itertools.chain(*examples[k])) for k in examples.keys()
        }
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        # We drop the small remainder, we could add padding if the model supported it instead of this drop, you can
        # customize this part to your needs.
        if total_length >= block_size:
            total_length = (total_length // block_size) * block_size
        # Split by chunks of max_len.
        result = {
            k: [t[i : i + block_size] for i in range(0, total_length, block_size)]
            for k, t in concatenated_examples.items()
        }
        result["labels"] = result["input_ids"].copy()
        return result

    lm_calib_set = tokenized_calib_set.map(
        group_texts,
        batched=True,
    )

    if n_blocks_for_stat > 0:
        assert len(lm_calib_set) > n_blocks_for_stat
        lm_calib_set = lm_calib_set.select(range(n_blocks_for_stat))

    calib_loader = torch.utils.data.DataLoader(
        lm_calib_set,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=True,
        drop_last=False,
        shuffle=False,
        collate_fn=default_data_collator,
    )

    return calib_loader



import os
from typing import Literal, Optional

from datasets import load_dataset as datasets_load_dataset

from fusion_bench.utils import validate_and_suggest_corrections

try:
    from modelscope import dataset_file_download as modelscope_dataset_file_download
    from modelscope import model_file_download as modelscope_model_file_download
    from modelscope import snapshot_download as modelscope_snapshot_download

except ImportError:

    def _raise_modelscope_not_installed_error(*args, **kwargs):
        raise ImportError(
            "ModelScope is not installed. Please install it using `pip install modelscope` to use ModelScope models."
        )

    modelscope_snapshot_download = _raise_modelscope_not_installed_error
    modelscope_model_file_download = _raise_modelscope_not_installed_error
    modelscope_dataset_file_download = _raise_modelscope_not_installed_error

try:
    from huggingface_hub import hf_hub_download
    from huggingface_hub import snapshot_download as huggingface_snapshot_download
except ImportError:

    def _raise_huggingface_not_installed_error(*args, **kwargs):
        raise ImportError(
            "Hugging Face Hub is not installed. Please install it using `pip install huggingface_hub` to use Hugging Face models."
        )

    huggingface_snapshot_download = _raise_huggingface_not_installed_error
    hf_hub_download = _raise_huggingface_not_installed_error

__all__ = [
    "load_dataset",
    "resolve_repo_path",
]

AVAILABLE_PLATFORMS = ["hf", "huggingface", "modelscope"]


def _raise_unknown_platform_error(platform: str):
    raise ValueError(
        f"Unsupported platform: {platform}. Supported platforms are 'hf', 'huggingface', and 'modelscope'."
    )


def load_dataset(
    name: str,
    split: str = "train",
    platform: Literal["hf", "huggingface", "modelscope"] = "hf",
):
    """
    Load a dataset from Hugging Face or ModelScope.

    Args:
        platform (Literal['hf', 'modelscope']): The platform to load the dataset from.
        name (str): The name of the dataset.
        split (str): The split of the dataset to load (default is "train").

    Returns:
        Dataset: The loaded dataset.
    """
    validate_and_suggest_corrections(platform, AVAILABLE_PLATFORMS)
    if platform == "hf" or platform == "huggingface":
        return datasets_load_dataset(name, split=split)
    elif platform == "modelscope":
        dataset_dir = modelscope_snapshot_download(name, repo_type="dataset")
        return datasets_load_dataset(dataset_dir, split=split)
    else:
        _raise_unknown_platform_error(platform)


def resolve_repo_path(
    repo_id: str,
    repo_type: Optional[str] = "model",
    platform: Literal["hf", "huggingface", "modelscope"] = "hf",
    **kwargs,
):
    """
    Resolve and download a repository from various platforms to a local path.

    This function handles multiple repository sources including local paths, Hugging Face,
    and ModelScope. It automatically downloads remote repositories to local cache and
    returns the local path for further use.

    Args:
        repo_id (str): Repository identifier. Can be:
            - Local file/directory path (returned as-is if exists)
            - Hugging Face model/dataset ID (e.g., "bert-base-uncased")
            - ModelScope model/dataset ID
            - URL-prefixed ID (e.g., "hf://model-name", "modelscope://model-name").
              The prefix will override the platform argument.
        repo_type (str, optional): Type of repository to download. Defaults to "model".
            Common values include "model" and "dataset".
        platform (Literal["hf", "huggingface", "modelscope"], optional):
            Platform to download from. Defaults to "hf". Options:
            - "hf" or "huggingface": Hugging Face Hub
            - "modelscope": ModelScope platform
        **kwargs: Additional arguments passed to the underlying download functions.

    Returns:
        str: Local path to the repository (either existing local path or downloaded cache path).

    Raises:
        FileNotFoundError: If the repository cannot be found or downloaded from any platform.
        ValueError: If an unsupported platform is specified.
        ImportError: If required dependencies for the specified platform are not installed.

    Examples:
        >>> # Local path (returned as-is)
        >>> resolve_repo_path("/path/to/local/model")
        "/path/to/local/model"

        >>> # Hugging Face model
        >>> resolve_repo_path("bert-base-uncased")
        "/home/user/.cache/huggingface/hub/models--bert-base-uncased/..."

        >>> # ModelScope model with explicit platform
        >>> resolve_repo_path("damo/nlp_bert_backbone_base_std", platform="modelscope")
        "/home/user/.cache/modelscope/hub/damo/nlp_bert_backbone_base_std/..."

        >>> # URL-prefixed repository ID
        >>> resolve_repo_path("hf://microsoft/DialoGPT-medium")
        "/home/user/.cache/huggingface/hub/models--microsoft--DialoGPT-medium/..."
    """
    # If it's a HuggingFace Hub model id, download snapshot
    if repo_id.startswith("hf://") or repo_id.startswith("huggingface://"):
        repo_id = repo_id.replace("hf://", "").replace("huggingface://", "")
        platform = "hf"
    # If it's a ModelScope model id, download snapshot
    elif repo_id.startswith("modelscope://"):
        repo_id = repo_id.replace("modelscope://", "")
        platform = "modelscope"

    # If it's a local file or directory, return as is
    if os.path.exists(repo_id):
        return repo_id

    try:
        validate_and_suggest_corrections(platform, AVAILABLE_PLATFORMS)
        # This will download the model to the cache and return the local path
        if platform in ["hf", "huggingface"]:
            local_path = huggingface_snapshot_download(
                repo_id=repo_id, repo_type=repo_type, **kwargs
            )
        elif platform == "modelscope":
            local_path = modelscope_snapshot_download(
                repo_id=repo_id, repo_type=repo_type, **kwargs
            )
        else:
            _raise_unknown_platform_error(platform)
        return local_path
    except Exception as e:
        raise FileNotFoundError(f"Could not resolve checkpoint: {repo_id}. Error: {e}")


def resolve_file_path(
    repo_id: str,
    filename: str,
    repo_type: Literal["model", "dataset"] = "model",
    platform: Literal["hf", "huggingface", "modelscope"] = "hf",
    **kwargs,
) -> str:
    """
    Resolve and download a specific file from a repository across multiple platforms.

    This function downloads a specific file from repositories hosted on various platforms
    including local paths, Hugging Face Hub, and ModelScope. It handles platform-specific
    URL prefixes and automatically determines the appropriate download method.

    Args:
        repo_id (str): Repository identifier. Can be:
            - Local directory path (file will be joined with this path if it exists)
            - Hugging Face model/dataset ID (e.g., "bert-base-uncased")
            - ModelScope model/dataset ID
            - URL-prefixed ID (e.g., "hf://model-name", "modelscope://model-name").
              The prefix will override the platform argument.
        filename (str): The specific file to download from the repository.
        repo_type (Literal["model", "dataset"], optional): Type of repository.
            Defaults to "model". Used for ModelScope platform to determine the
            correct download function.
        platform (Literal["hf", "huggingface", "modelscope"], optional):
            Platform to download from. Defaults to "hf". Options:
            - "hf" or "huggingface": Hugging Face Hub
            - "modelscope": ModelScope platform
        **kwargs: Additional arguments passed to the underlying download functions
            (e.g., cache_dir, force_download, use_auth_token).

    Returns:
        str: Local path to the downloaded file.

    Raises:
        ValueError: If an unsupported repo_type is specified for ModelScope platform.
        ImportError: If required dependencies for the specified platform are not installed.
        FileNotFoundError: If the file cannot be found or downloaded.

    Examples:
        >>> # Download config.json from a Hugging Face model
        >>> resolve_file_path("bert-base-uncased", "config.json")
        "/home/user/.cache/huggingface/hub/models--bert-base-uncased/.../config.json"

        >>> # Download from ModelScope
        >>> resolve_file_path(
        ...     "damo/nlp_bert_backbone_base_std",
        ...     "pytorch_model.bin",
        ...     platform="modelscope"
        ... )
        "/home/user/.cache/modelscope/hub/.../pytorch_model.bin"

        >>> # Local file path
        >>> resolve_file_path("/path/to/local/model", "config.json")
        "/path/to/local/model/config.json"

        >>> # URL-prefixed repository
        >>> resolve_file_path("hf://microsoft/DialoGPT-medium", "config.json")
        "/home/user/.cache/huggingface/hub/.../config.json"

        >>> # Download dataset file from ModelScope
        >>> resolve_file_path(
        ...     "DAMO_NLP/jd",
        ...     "train.json",
        ...     repo_type="dataset",
        ...     platform="modelscope"
        ... )
        "/home/user/.cache/modelscope/datasets/.../train.json"
    """
    # If it's a HuggingFace Hub model id, download snapshot
    if repo_id.startswith("hf://") or repo_id.startswith("huggingface://"):
        repo_id = repo_id.replace("hf://", "").replace("huggingface://", "")
        platform = "hf"
    # If it's a ModelScope model id, download snapshot
    elif repo_id.startswith("modelscope://"):
        repo_id = repo_id.replace("modelscope://", "")
        platform = "modelscope"

    # If it's a local file or directory, return as is
    if os.path.exists(repo_id):
        return os.path.join(repo_id, filename)

    if platform in ["hf", "huggingface"]:
        return hf_hub_download(
            repo_id=repo_id,
            filename=filename,
            repo_type=repo_type,
            **kwargs,
        )
    elif platform == "modelscope":
        if repo_type == "model":
            return modelscope_model_file_download(
                model_id=repo_id, file_path=filename, **kwargs
            )
        elif repo_type == "dataset":
            return modelscope_dataset_file_download(
                dataset_id=repo_id, file_path=filename, **kwargs
            )
        else:
            raise ValueError(
                f"Unsupported repo_type: {repo_type}. Supported types are 'model' and 'dataset'."
            )
    else:
        _raise_unknown_platform_error(platform)



from typing import Union

from omegaconf import DictConfig
from tqdm.autonotebook import tqdm


class TaskPool:
    """
    A class to manage a pool of tasks for evaluation.
    This is the base class for version 0.1.x, deprecated.
    Use `fusion_bench.taskpool.BaseTaskPool` instead.

    Attributes:
        config (DictConfig): The configuration for the task pool.
        _all_task_names (List[str]): A list of all task names in the task pool.
    """

    _program = None

    def __init__(self, taskpool_config: DictConfig):
        """
        Initialize the TaskPool with the given configuration.

        Args:
            taskpool_config (DictConfig): The configuration for the task pool.
        """
        super().__init__()
        self.config = taskpool_config

        # Check for duplicate task names
        if self.config.get("tasks", None) is not None:
            task_names = [task["name"] for task in self.config["tasks"]]
            assert len(task_names) == len(
                set(task_names)
            ), "Duplicate task names found in the task pool"
            self._all_task_names = task_names

    def evaluate(self, model):
        """
        Evaluate the model on all tasks in the task pool, and return a report.

        Take image classification as an example, the report will look like:

        ```python
        {
            "mnist": {
                "accuracy": 0.8,
                "loss": 0.2,
            },
            <task_name>: {
                <metric_name>: <metric_value>,
                ...
            },
        }
        ```

        Args:
            model: The model to evaluate.

        Returns:
            report (dict): A dictionary containing the results of the evaluation for each task.
        """
        report = {}
        for task_name in tqdm(self.task_names, desc="Evaluating tasks"):
            task = self.load_task(task_name)
            result = task.evaluate(model)
            report[task_name] = result
        return report

    @property
    def task_names(self):
        """
        Return a list of all task names in the task pool.

        Returns:
            List[str]: A list of all task names.
        """
        return self._all_task_names

    def get_task_config(self, task_name: str):
        """
        Retrieve the configuration for a specific task from the task pool.

        Args:
            task_name (str): The name of the task for which to retrieve the configuration.

        Returns:
            DictConfig: The configuration dictionary for the specified task.

        Raises:
            ValueError: If the specified task is not found in the task pool.
        """
        for task in self.config["tasks"]:
            if task["name"] == task_name:
                return task
        raise ValueError(f"Task {task_name} not found in the task pool")

    def load_task(self, task_name_or_config: Union[str, DictConfig]):
        """
        Load a task from the task pool.

        Args:
            task_name_or_config (Union[str, DictConfig]): The name or configuration of the task to load.

        Returns:
            Any: The loaded task.

        Raises:
            NotImplementedError: If the method is not implemented in the subclass.
        """
        raise NotImplementedError



from typing import Dict, Tuple

import torch
from torch import Tensor, nn

from fusion_bench.models.modeling_deepseek_v2 import DeepseekV2MoEGate

from .hook import BaseHookFn


class MoEPrunerHookFnForDeepseekV2Linear(BaseHookFn):
    _routing_weights = None  # set by gate hook

    def __init__(self, linear: nn.Linear, name: str):
        super().__init__(linear)
        self.linear = linear
        self.scalar_row = torch.zeros(
            (linear.weight.size(1),), device=linear.weight.device
        )
        self.nsamples = 0
        self.name = name

    def __call__(self, linear, inps: Tuple[Tensor], out: Tensor):
        assert len(inps) == 1
        inp = inps[0]
        if len(inp.shape) == 2:
            inp = inp.unsqueeze(0)

        batch_size = inp.shape[0]
        if len(inp.shape) == 3:
            inp = inp.reshape((-1, inp.shape[-1]))
        # (NxL, C) -> (C, NxL)
        inp = inp.t()

        self.scalar_row *= self.nsamples / (self.nsamples + batch_size)
        self.nsamples += batch_size

        inp = inp.type(torch.float32)
        routing_weights = self._routing_weights.t()
        self.scalar_row += (
            torch.norm(inp * routing_weights, p=2, dim=1) ** 2 / self.nsamples
        )

    def compute(self):
        return torch.abs(self.linear.weight) * torch.sqrt(
            self.scalar_row.reshape(1, -1)
        )


class MoEPrunerHookFnForDeepseekV2Gate(BaseHookFn):
    def __init__(
        self,
        router: DeepseekV2MoEGate,
        linear_layer_hooks: Dict[str, MoEPrunerHookFnForDeepseekV2Linear],
        top_k: int,
        num_experts: int,
    ):
        super().__init__(router)
        self.router = router
        self.linear_layer_hooks = linear_layer_hooks
        self.top_k = top_k
        self.num_experts = num_experts

    def __call__(self, router, inps: Tuple[Tensor], out: Tuple[Tensor, Tensor, Tensor]):
        assert len(inps) == 1

        x = inps[0]
        x = x.view(-1, x.shape[-1])
        topk_ids, topk_weight, aux_loss = out

        # One hot encode the selected experts to create an expert mask
        # this will be used to easily index which expert is going to be sollicitated
        expert_mask = torch.nn.functional.one_hot(
            topk_ids, num_classes=self.num_experts
        ).permute(2, 1, 0)

        for expert_idx in range(self.num_experts):
            idx, top_x = torch.where(expert_mask[expert_idx])
            for name, hook in self.linear_layer_hooks.items():
                if not name.startswith(f"{expert_idx}."):
                    continue
                hook._routing_weights = topk_weight[top_x, idx, None]

    def compute(self):
        pass



import threading
from pathlib import Path
from typing import Optional, Union


class RuntimeConstants:
    """
    This class holds constants related to the runtime environment of the Fusion Bench framework.
    It includes default values for cache directories and other runtime configurations.

    Implemented as a thread-safe singleton to ensure consistent runtime configuration
    across the entire application.
    """

    _instance: Optional["RuntimeConstants"] = None
    _lock = threading.Lock()

    def __new__(cls) -> "RuntimeConstants":
        """Create a new instance using singleton pattern with thread safety."""
        with cls._lock:
            # Double-check locking pattern
            if cls._instance is None:
                cls._instance = super(RuntimeConstants, cls).__new__(cls)
                cls._instance._initialized = False
            return cls._instance

    def __init__(self):
        """Initialize the singleton instance only once."""
        if not self._initialized:
            # Add your runtime constants here
            self._initialized = True

    debug = False

    @property
    def cache_dir(self) -> Path:
        from fusion_bench.utils.cache_utils import DEFAULT_CACHE_DIR

        return DEFAULT_CACHE_DIR

    @cache_dir.setter
    def cache_dir(self, path: Union[str, Path]) -> None:
        from fusion_bench.utils.cache_utils import set_default_cache_dir

        set_default_cache_dir(path)

    @property
    def print_function_call(self) -> bool:
        from fusion_bench.utils.instantiate_utils import PRINT_FUNCTION_CALL

        return PRINT_FUNCTION_CALL

    @print_function_call.setter
    def print_function_call(self, enable: bool) -> None:
        from fusion_bench.utils.instantiate_utils import set_print_function_call

        set_print_function_call(enable)
